from __future__ import annotations

import math
import re
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, Literal, NoReturn, cast

import xdsl.parser as affine_parser
from xdsl.context import Context
from xdsl.dialects.builtin import (
    AffineMapAttr,
    AffineSetAttr,
    AnyDenseElement,
    AnyFloat,
    AnyFloatConstr,
    AnyTensorType,
    AnyUnrankedTensorType,
    AnyVectorType,
    ArrayAttr,
    BFloat16Type,
    BoolAttr,
    BytesAttr,
    ComplexType,
    DenseArrayBase,
    DenseIntOrFPElementsAttr,
    DenseResourceAttr,
    DictionaryAttr,
    Float16Type,
    Float32Type,
    Float64Type,
    Float80Type,
    Float128Type,
    FloatAttr,
    FunctionType,
    IndexType,
    IntegerAttr,
    IntegerType,
    LocationAttr,
    MemRefLayoutAttr,
    MemRefType,
    NoneAttr,
    NoneType,
    OpaqueAttr,
    RankedStructure,
    Signedness,
    StridedLayoutAttr,
    StringAttr,
    SymbolRefAttr,
    TensorType,
    UnitAttr,
    UnrankedMemRefType,
    UnrankedTensorType,
    UnregisteredAttr,
    VectorType,
    i64,
)
from xdsl.ir import Attribute, Data, ParametrizedAttribute
from xdsl.ir.affine import AffineMap, AffineSet
from xdsl.irdl import base
from xdsl.utils.bitwise_casts import (
    convert_u16_to_f16,
    convert_u32_to_f32,
    convert_u64_to_f64,
)
from xdsl.utils.exceptions import ParseError, VerifyException
from xdsl.utils.isattr import isattr
from xdsl.utils.lexer import Position, Span
from xdsl.utils.mlir_lexer import MLIRTokenKind, StringLiteral

from .base_parser import BaseParser  # noqa: TID251


@dataclass
class AttrParser(BaseParser):
    """
    Basic recursive descent parser for attributes and types.

    Methods named `parse_*` will consume tokens, and throw a `ParseError` if
    an unexpected token is parsed. Methods marked with `parse_optional` will return
    None if the first token is unexpected, and will throw a `ParseError` if the
    first token is expected, but a following token is not.

    Methods with a `context_msg` argument allows to append the context message to the
    thrown error. For instance, if `',' expected` is returned, setting `context_msg` to
    `" in integer list"` will throw the error `',' expected in integer list` instead.

    """

    ctx: Context

    attribute_aliases: dict[str, Attribute] = field(default_factory=dict)
    """
    A dictionary of aliases for attributes.
    The key is the alias name, including the `!` or `#` prefix.
    """

    def parse_optional_type(self) -> Attribute | None:
        """
        Parse an xDSL type, if present.
        An xDSL type is either a builtin type, which can have various format,
        or a dialect type, with the following format:
            type          ::= builtin-type | dialect-type | alias-type
            alias-type    ::= `!` type-name
            dialect-type  ::= `!` type-name (`<` dialect-type-contents+ `>`)?
            type-name     ::= bare-id
            dialect-type-contents ::= `<` dialect-attribute-contents+ `>`
                            | `(` dialect-attribute-contents+ `)`
                            | `[` dialect-attribute-contents+ `]`
                            | `{` dialect-attribute-contents+ `}`
                            | [^[]<>(){}\0]+
        """
        if (
            token := self._parse_optional_token(MLIRTokenKind.EXCLAMATION_IDENT)
        ) is not None:
            return self._parse_extended_type_or_attribute(token.text[1:], True)
        return self._parse_optional_builtin_type()

    def parse_type(self) -> Attribute:
        """
        Parse an xDSL type.
        An xDSL type is either a builtin type, which can have various format,
        or a dialect or alias type, with the following format:
            type          ::= builtin-type | dialect-type | alias-type
            alias-type    ::= `!` type-name
            dialect-type  ::= `!` type-name (`<` dialect-type-contents+ `>`)?
            type-name     ::= bare-id
            dialect-type-contents ::= `<` dialect-attribute-contents+ `>`
                            | `(` dialect-attribute-contents+ `)`
                            | `[` dialect-attribute-contents+ `]`
                            | `{` dialect-attribute-contents+ `}`
                            | [^[]<>(){}\0]+
        """
        return self.expect(self.parse_optional_type, "type expected")

    def parse_optional_attribute(self) -> Attribute | None:
        """
        Parse an xDSL attribute, if present.
        An attribute is either a builtin attribute, which can have various format,
        or a dialect or alias attribute, with the following format:
            attr          ::= builtin-attr | dialect-attr | alias-attr
            alias-attr    ::= `!` attr-name
            dialect-attr  ::= `#` attr-name (`<` dialect-attr-contents+ `>`)?
            attr-name     ::= bare-id
            dialect-attr-contents ::= `<` dialect-attribute-contents+ `>`
                            | `(` dialect-attribute-contents+ `)`
                            | `[` dialect-attribute-contents+ `]`
                            | `{` dialect-attribute-contents+ `}`
                            | [^[]<>(){}\0]+
        """
        if (token := self._parse_optional_token(MLIRTokenKind.HASH_IDENT)) is not None:
            return self._parse_extended_type_or_attribute(token.text[1:], False)
        return self._parse_optional_builtin_attr()

    def parse_attribute(self) -> Attribute:
        """
        Parse an xDSL attribute.
        An attribute is either a builtin attribute, which can have various format,
        or a dialect attribute, with the following format:
            attr          ::= builtin-attr | dialect-attr | alias-attr
            alias-attr    ::= `!` attr-name
            dialect-attr  ::= `#` attr-name (`<` dialect-attr-contents+ `>`)?
            attr-name     ::= bare-id
            dialect-attr-contents ::= `<` dialect-attribute-contents+ `>`
                            | `(` dialect-attribute-contents+ `)`
                            | `[` dialect-attribute-contents+ `]`
                            | `{` dialect-attribute-contents+ `}`
                            | [^[]<>(){}\0]+
        """
        return self.expect(self.parse_optional_attribute, "attribute expected")

    def _parse_attribute_entry(self) -> tuple[str, Attribute]:
        """
        Parse entry in attribute dict. Of format:

        attribute_entry := (bare-id | string-literal) `=` attribute
        attribute       := dialect-attribute | builtin-attribute
        """
        if (name := self._parse_optional_token(MLIRTokenKind.BARE_IDENT)) is not None:
            name = name.span.text
        else:
            name = self.parse_optional_str_literal()

        if name is None:
            self.raise_error(
                "Expected bare-id or string-literal here as part of attribute entry!"
            )

        if self.parse_optional_punctuation("=") is None:
            return name, UnitAttr()

        return name, self.parse_attribute()

    def _find_duplicated_key(self, attrs: list[tuple[str, Attribute]]) -> str | None:
        seen_keys: set[str] = set()
        for key, _ in attrs:
            if key in seen_keys:
                return key
            seen_keys.add(key)
        return None

    def parse_optional_dictionary_attr_dict(self) -> dict[str, Attribute]:
        attrs = self.parse_optional_comma_separated_list(
            self.Delimiter.BRACES, self._parse_attribute_entry
        )
        if attrs is None:
            return dict()

        if (key := self._find_duplicated_key(attrs)) is not None:
            self.raise_error(f"Duplicate key '{key}' in dictionary attribute")

        return dict(attrs)

    def _parse_dialect_type_or_attribute_body(
        self,
        attr_name: str,
        is_type: bool,
        is_opaque: bool,
        starting_opaque_pos: Position | None,
    ):
        """
        Parse the contents of an attribute or type, with syntax:
            dialect-attr-contents ::= `<` dialect-attr-contents+ `>`
                                    | `(` dialect-attr-contents+ `)`
                                    | `[` dialect-attr-contents+ `]`
                                    | `{` dialect-attr-contents+ `}`
                                    | [^[]<>(){}\0]+
        In the case where the attribute or type is using the opaque syntax,
        the attribute or type mnemonic should have already been parsed.
        """
        pretty = "." in attr_name
        if not pretty:
            self.parse_punctuation("<")
            attr_name += (
                "."
                + self._parse_token(
                    MLIRTokenKind.BARE_IDENT, "Expected attribute name."
                ).text
            )
        attr_def = self.ctx.get_optional_attr(
            attr_name,
            create_unregistered_as_type=is_type,
        )
        if attr_def is None:
            self.raise_error(f"'{attr_name}' is not registered")
        if issubclass(attr_def, UnregisteredAttr):
            if not is_opaque:
                if self.parse_optional_punctuation("<") is None:
                    return attr_def(attr_name, is_type, is_opaque, "")
            body = self._parse_unregistered_attr_body(starting_opaque_pos)
            attr = attr_def(attr_name, is_type, is_opaque, body)
            if not is_opaque:
                self.parse_punctuation(">")
            return attr

        elif issubclass(attr_def, ParametrizedAttribute):
            param_list = attr_def.parse_parameters(self)
            return attr_def.new(param_list)
        elif issubclass(attr_def, Data):
            param: Any = attr_def.parse_parameter(self)
            return cast(Data[Any], attr_def(param))
        else:
            raise TypeError("Attributes are either ParametrizedAttribute or Data.")

    def _parse_extended_type_or_attribute(
        self, attr_or_dialect_name: str, is_type: bool = True
    ) -> Attribute:
        """
        Parse the contents of a dialect or alias type or attribute, with format:
            dialect-attr-contents ::= `<` dialect-attr-contents+ `>`
                                    | `(` dialect-attr-contents+ `)`
                                    | `[` dialect-attr-contents+ `]`
                                    | `{` dialect-attr-contents+ `}`
                                    | [^[]<>(){}\0]+
        The contents will be parsed by a user-defined parser, or by a generic parser
        if the dialect attribute/type is not registered.

        In the case that the type or attribute is using the opaque syntax (where the
        identifier parsed is the dialect name), this function will parse the opaque
        attribute with the following format:
            opaque-attr-contents ::= `<` bare-ident dialect-attr-contents+ `>`
        otherwise, it will parse them with the pretty or alias syntax, with format:
            pretty-or-alias-attr-contents ::= `<` dialect-attr-contents+ `>`
        """
        is_pretty_name = "." in attr_or_dialect_name
        starting_opaque_pos = None

        if not is_pretty_name:
            # An attribute or type alias
            if self.parse_optional_punctuation("<") is None:
                alias_name = ("!" if is_type else "#") + attr_or_dialect_name
                if alias_name not in self.attribute_aliases:
                    self.raise_error(f"undefined symbol alias '{alias_name}'")
                return self.attribute_aliases[alias_name]

            # An opaque dialect attribute or type
            # Compared to MLIR, we still go through the symbol parser, instead of the
            # dialect parser.
            if not is_pretty_name:
                attr_name_token = self._parse_token(
                    MLIRTokenKind.BARE_IDENT, "Expected attribute name."
                )
                starting_opaque_pos = attr_name_token.span.end

                attr_or_dialect_name += "." + attr_name_token.text

        attr = self._parse_dialect_type_or_attribute_body(
            attr_or_dialect_name, is_type, not is_pretty_name, starting_opaque_pos
        )

        if not is_pretty_name:
            self.parse_punctuation(">")

        return attr

    def _parse_unregistered_attr_body(self, start_pos: Position | None) -> str:
        """
        Parse the body of an unregistered attribute, which is a balanced
        string for `<`, `(`, `[`, `{`, and may contain string literals.
        The body ends when no parentheses are opened, and an `>` is encountered.
        """

        if start_pos is None:
            start_pos = self.pos
        end_pos: Position = start_pos

        symbols_stack: list[MLIRTokenKind] = []
        parentheses = {
            MLIRTokenKind.GREATER: MLIRTokenKind.LESS,
            MLIRTokenKind.R_PAREN: MLIRTokenKind.L_PAREN,
            MLIRTokenKind.R_SQUARE: MLIRTokenKind.L_SQUARE,
            MLIRTokenKind.R_BRACE: MLIRTokenKind.L_BRACE,
        }
        parentheses_names = {
            MLIRTokenKind.GREATER: "`>`",
            MLIRTokenKind.R_PAREN: "`)`",
            MLIRTokenKind.R_SQUARE: "`]`",
            MLIRTokenKind.R_BRACE: "`}`",
        }
        while True:
            # Opening a new parenthesis
            if (
                token := self._parse_optional_token_in(parentheses.values())
            ) is not None:
                symbols_stack.append(token.kind)
                continue

            # Closing a parenthesis
            if (token := self._current_token).kind in parentheses.keys():
                closing = parentheses[token.kind]

                # If we don't have any open parenthesis, either we end the parsing if
                # the parenthesis is a `>`, or we raise an error.
                if len(symbols_stack) == 0:
                    if token.kind == MLIRTokenKind.GREATER:
                        end_pos = self.pos
                        break
                    self.raise_error(
                        "Unexpected closing parenthesis "
                        f"{parentheses_names[token.kind]} in attribute body!",
                        self._current_token.span,
                    )

                # If we have an open parenthesis, check that we are closing it
                # with the right parenthesis kind.
                if symbols_stack[-1] != closing:
                    self.raise_error(
                        "Unexpected closing parenthesis "
                        f"{parentheses_names[token.kind]} in attribute body! {symbols_stack}",
                        self._current_token.span,
                    )
                symbols_stack.pop()
                self._consume_token()
                continue

            # Checking for unexpected EOF
            if self._parse_optional_token(MLIRTokenKind.EOF) is not None:
                self.raise_error(
                    "Unexpected end of file before closing of attribute body!"
                )

            # Other tokens
            self._consume_token()

        body = self.lexer.input.slice(start_pos, end_pos)
        assert body is not None
        return body

    def _parse_optional_builtin_parametrized_type(self) -> ParametrizedAttribute | None:
        """
        Parse an builtin parametrized type, if present, with format:
            builtin-parametrized-type ::= builtin-name `<` args `>`
            builtin-name ::= vector | memref | tensor | complex | tuple
            args ::= <defined by the builtin name>
        """
        if self._current_token.kind != MLIRTokenKind.BARE_IDENT:
            return None

        name = self._current_token.text

        def unimplemented() -> NoReturn:
            raise ParseError(
                self._current_token.span,
                f"Builtin {name} is not supported yet!",
            )

        builtin_parsers: dict[str, Callable[[], ParametrizedAttribute]] = {
            "vector": self._parse_vector_attrs,
            "memref": self._parse_memref_attrs,
            "tensor": self._parse_tensor_attrs,
            "complex": self._parse_complex_attrs,
            "tuple": unimplemented,
        }

        if name not in builtin_parsers:
            return None
        self._consume_token(MLIRTokenKind.BARE_IDENT)

        self.parse_punctuation("<", " after builtin name")
        # Get the parser for the type, falling back to the unimplemented warning
        res = builtin_parsers.get(name, unimplemented)()
        self.parse_punctuation(">", " after builtin parameter list")
        return res

    def parse_shape_dimension(self, allow_dynamic: bool = True) -> int:
        """
        Parse a single shape dimension, which is a decimal literal or `?`.
        `?` is interpreted as -1. Note that if the integer literal is in
        hexadecimal form, it will be split into multiple tokens. For example,
        `0x10` will be split into `0` and `x10`.
        Optionally allows to not parse `?` as -1.
        """
        if self._current_token.kind not in (
            MLIRTokenKind.INTEGER_LIT,
            MLIRTokenKind.QUESTION,
        ):
            if allow_dynamic:
                self.raise_error(
                    "Expected either integer literal or '?' in shape dimension, "
                    f"got {self._current_token.kind.name}!"
                )
            self.raise_error(
                "Expected integer literal in shape dimension, "
                f"got {self._current_token.kind.name}!"
            )

        if self.parse_optional_punctuation("?") is not None:
            if allow_dynamic:
                return -1
            self.raise_error("Unexpected dynamic dimension!")

        # If the integer literal starts with `0x`, this is decomposed into
        # `0` and `x`.
        int_token = self._consume_token(MLIRTokenKind.INTEGER_LIT)
        if int_token.text[:2] == "0x":
            self._resume_from(int_token.span.start + 1)
            return 0

        return int_token.kind.get_int_value(int_token.span)

    def parse_shape_delimiter(self) -> None:
        """
        Parse 'x', a shape delimiter. Note that if 'x' is followed by other
        characters, it will split the token. For instance, 'x1' will be split
        into 'x' and '1'.
        """
        if self._current_token.kind != MLIRTokenKind.BARE_IDENT:
            self.raise_error(
                f"Expected 'x' in shape delimiter, got {self._current_token.kind.name}"
            )

        if self._current_token.text[0] != "x":
            self.raise_error(
                f"Expected 'x' in shape delimiter, got {self._current_token.text}"
            )

        # Move the lexer to the position after 'x'.
        self._resume_from(self._current_token.span.start + 1)

    def parse_ranked_shape(self) -> tuple[list[int], Attribute]:
        """
        Parse a ranked shape with the following format:
          ranked-shape ::= (dimension `x`)* type
          dimension ::= `?` | decimal-literal
        each dimension is also required to be non-negative.
        """
        dims: list[int] = []
        while self._current_token.kind in (
            MLIRTokenKind.INTEGER_LIT,
            MLIRTokenKind.QUESTION,
        ):
            dim = self.parse_shape_dimension()
            dims.append(dim)
            self.parse_shape_delimiter()

        type = self.expect(self.parse_optional_type, "Expected shape type.")
        return dims, type

    def parse_shape(self) -> tuple[list[int] | None, Attribute]:
        """
        Parse a ranked or unranked shape with the following format:

        shape ::= ranked-shape | unranked-shape
        ranked-shape ::= (dimension `x`)* type
        unranked-shape ::= `*`x type
        dimension ::= `?` | decimal-literal

        each dimension is also required to be non-negative.
        """
        if self.parse_optional_punctuation("*") is not None:
            self.parse_shape_delimiter()
            type = self.expect(self.parse_optional_type, "Expected shape type.")
            return None, type
        return self.parse_ranked_shape()

    def _parse_complex_attrs(self) -> ComplexType:
        element_type = self.parse_attribute()
        if not isattr(element_type, base(IntegerType) | AnyFloatConstr):
            self.raise_error(
                "Complex type must be parameterized by an integer or float type!"
            )
        return ComplexType(element_type)

    def _parse_memref_attrs(
        self,
    ) -> MemRefType[Attribute] | UnrankedMemRefType[Attribute]:
        shape, type = self.parse_shape()

        # Unranked case
        if shape is None:
            if self.parse_optional_punctuation(",") is None:
                return UnrankedMemRefType.from_type(type)
            memory_space = self.parse_attribute()
            return UnrankedMemRefType.from_type(type, memory_space)

        if self.parse_optional_punctuation(",") is None:
            return MemRefType(type, shape)

        memory_or_layout = self.parse_attribute()

        # If there is both a memory space and a layout, we know that the
        # layout is the second one
        if self.parse_optional_punctuation(",") is not None:
            memory_space = self.parse_attribute()
            if not isinstance(memory_or_layout, MemRefLayoutAttr):
                self.raise_error("Expected a MemRef layout attribute")
            return MemRefType(type, shape, memory_or_layout, memory_space)

        # If the argument is a MemRefLayoutAttr, use it as layout
        if isinstance(memory_or_layout, MemRefLayoutAttr):
            return MemRefType(type, shape, layout=memory_or_layout)

        # Otherwise, consider it as the memory space.
        else:
            return MemRefType(type, shape, memory_space=memory_or_layout)

    def _parse_vector_attrs(self) -> AnyVectorType:
        dims: list[int] = []
        scalable_dims: list[bool] = []

        while True:
            if self._current_token.kind == MLIRTokenKind.INTEGER_LIT:
                # Static dimension
                dims.append(self.parse_shape_dimension(allow_dynamic=False))
                scalable_dims.append(False)
                self.parse_shape_delimiter()
            elif self.parse_optional_punctuation("[") is not None:
                # Scalable dimension
                dims.append(self.parse_shape_dimension(allow_dynamic=False))
                scalable_dims.append(True)
                self.parse_punctuation("]")
                self.parse_shape_delimiter()
            else:
                break

        type = self.parse_optional_type()
        if type is None:
            self.raise_error("Expected vector element type")

        scalable_dims_attr = ArrayAttr(BoolAttr.from_bool(s) for s in scalable_dims)

        return VectorType(type, dims, scalable_dims_attr)

    def _parse_tensor_attrs(self) -> AnyTensorType | AnyUnrankedTensorType:
        shape, type = self.parse_shape()

        if shape is None:
            if self.parse_optional_punctuation(",") is not None:
                self.raise_error("Unranked tensors don't have an encoding!")
            return UnrankedTensorType(type)

        if self.parse_optional_punctuation(",") is not None:
            encoding = self.parse_attribute()
            return TensorType(type, shape, encoding)

        return TensorType(type, shape)

    def _parse_attribute_type(self) -> Attribute:
        """
        Parses `:` type and returns the type
        """
        self.parse_characters(":", " in attribute type")
        return self.parse_type()

    def _parse_optional_builtin_attr(self) -> Attribute | None:
        """
        Tries to parse a builtin attribute, e.g. a string literal, int, array, etc..
        """

        # String literal
        if (str_lit := self.parse_optional_str_literal()) is not None:
            return StringAttr(str_lit)
        # Bytes literal
        if (bytes_lit := self.parse_optional_bytes_literal()) is not None:
            return BytesAttr(bytes_lit)

        attrs = (
            self.parse_optional_unit_attr,
            self.parse_optional_builtin_int_or_float_attr,
            self._parse_optional_array_attr,
            self._parse_optional_symref_attr,
            self.parse_optional_location,
            self._parse_optional_builtin_dict_attr,
            self.parse_optional_type,
            self._parse_optional_builtin_parametrized_attr,
        )

        for attr_parser in attrs:
            if (val := attr_parser()) is not None:
                return val

        return None

    def _parse_int_or_question(self, context_msg: str = "") -> int | Literal["?"]:
        """Parse either an integer literal, or a '?'."""
        if self._parse_optional_token(MLIRTokenKind.QUESTION) is not None:
            return "?"
        if (v := self.parse_optional_integer(allow_boolean=False)) is not None:
            return v
        self.raise_error("Expected an integer literal or `?`" + context_msg)

    def _parse_strided_layout_attr(self) -> Attribute:
        """
        Parse a strided layout attribute parameters.
        | `<` `[` comma-separated-int-or-question `]`
          (`,` `offset` `:` integer-literal)? `>`
        """
        # Parse stride list
        self._parse_token(MLIRTokenKind.LESS, "Expected `<` after `strided`")
        strides = self.parse_comma_separated_list(
            self.Delimiter.SQUARE,
            lambda: self._parse_int_or_question(" in stride list"),
            " in stride list",
        )
        # Pyright widen `Literal['?']` to `str` for some reasons
        strides = cast(list[int | Literal["?"]], strides)

        # Convert to the attribute expected input
        strides = [None if stride == "?" else stride for stride in strides]

        # Case without offset
        if self._parse_optional_token(MLIRTokenKind.GREATER) is not None:
            return StridedLayoutAttr(strides)

        # Parse the optional offset
        self._parse_token(
            MLIRTokenKind.COMMA, "Expected end of strided attribute or ',' for offset."
        )
        self.parse_keyword("offset", " after comma")
        self._parse_token(MLIRTokenKind.COLON, "Expected ':' after 'offset'")
        offset = self._parse_int_or_question(" in stride offset")
        self._parse_token(
            MLIRTokenKind.GREATER, "Expected '>' in end of stride attribute"
        )
        return StridedLayoutAttr(strides, None if offset == "?" else offset)

    def parse_optional_unit_attr(self) -> Attribute | None:
        """
        Parse a value of `unit` type.
        unit-attribute ::= `unit`
        """
        if self._current_token.kind != MLIRTokenKind.BARE_IDENT:
            return None
        name = self._current_token.span.text

        # Unit attribute
        if name == "unit":
            self._consume_token()
            return UnitAttr()

        return None

    def _parse_optional_builtin_parametrized_attr(self) -> Attribute | None:
        if self._current_token.kind != MLIRTokenKind.BARE_IDENT:
            return None
        name = self._current_token.span
        parsers = {
            "dense": self._parse_builtin_dense_attr,
            "opaque": self._parse_builtin_opaque_attr,
            "dense_resource": self._parse_builtin_dense_resource_attr,
            "array": self._parse_builtin_densearray_attr,
            "affine_map": self._parse_builtin_affine_map,
            "affine_set": self._parse_builtin_affine_set,
            "strided": self._parse_strided_layout_attr,
        }

        if name.text not in parsers:
            return None
        self._consume_token(MLIRTokenKind.BARE_IDENT)
        return parsers[name.text]()

    def _parse_dense_literal_type(
        self,
    ) -> (
        RankedStructure[IntegerType]
        | RankedStructure[IndexType]
        | RankedStructure[AnyFloat]
    ):
        type = self.expect(self.parse_optional_type, "Dense attribute must be typed!")
        # Check that the type is correct.
        if not isattr(
            type,
            base(RankedStructure[IntegerType])
            | base(RankedStructure[IndexType])
            | base(RankedStructure[AnyFloat]),
        ):
            self.raise_error(
                "Expected memref, vector or tensor type of "
                "integer, index, or float type"
            )

        # Check for static shapes in type
        if any(dim == -1 for dim in list(type.get_shape())):
            self.raise_error("Dense literal attribute should have a static shape.")
        return type

    def parse_dense_int_or_fp_elements_attr(
        self, type: RankedStructure[AnyDenseElement] | None
    ) -> DenseIntOrFPElementsAttr:
        dense_contents: (
            tuple[list[AttrParser._TensorLiteralElement], list[int]] | str | None
        )
        """
        If `None`, then the contents are empty.
        If `str`, then this is a hex-encoded string containing the data, which doesn't
        carry shape information.
        Otherwise, a tuple of `elements` and `shape`.
        If `shape` is `[]`, then this is a splat attribute, meaning it has the same value
        everywhere.
        """

        self.parse_punctuation("<", " in dense attribute")
        if self.parse_optional_punctuation(">") is not None:
            # Empty case
            dense_contents = None
        else:
            if (hex_string := self.parse_optional_str_literal()) is not None:
                dense_contents, shape = hex_string, None
            else:
                # Expect a tensor literal instead
                dense_contents = self._parse_tensor_literal()
            self.parse_punctuation(">", " in dense attribute")

        # Parse the dense type and check for correctness
        if type is None:
            self.parse_punctuation(":", " in dense attribute")
            type = self._parse_dense_literal_type()
        type_shape = list(type.get_shape())
        type_num_values = math.prod(type_shape)

        if dense_contents is None:
            # Empty case
            if type_num_values != 0:
                self.raise_error(
                    "Expected at least one element in the dense literal, but got None"
                )
            data_values = []
        elif isinstance(dense_contents, str):
            # Hex-encoded string case: convert straight to bytes (without the 0x prefix)
            try:
                bytes_values = bytes.fromhex(dense_contents[2:])
            except ValueError:
                self.raise_error("Hex string in denseAttr is invalid")

            # Handle splat values given in hex
            if len(bytes_values) == type.element_type.compile_time_size:
                bytes_values *= type_num_values

            # Create attribute
            attr = DenseIntOrFPElementsAttr([type, BytesAttr(bytes_values)])
            if type_num_values != len(attr):
                self.raise_error(
                    f"Shape mismatch in dense literal. Expected {type_num_values} "
                    f"elements from the type, but got {len(attr)} elements."
                )
            return attr

        else:
            # Tensor literal case
            dense_values, shape = dense_contents
            data_values = [
                value.to_type(self, type.element_type) for value in dense_values
            ]
            # Elements from _parse_tensor_literal need to be converted to values.
            if shape:
                # Check that the shape matches the data when given a shaped data.
                # For splat attributes any shape is fine
                if type_shape != shape:
                    self.raise_error(
                        f"Shape mismatch in dense literal. Expected {type_shape} "
                        f"shape from the type, but got {shape} shape."
                    )
            else:
                assert len(data_values) == 1, "Fatal error in parser"
                data_values *= type_num_values

        return DenseIntOrFPElementsAttr.from_list(type, data_values)

    def _parse_builtin_dense_attr(self) -> DenseIntOrFPElementsAttr:
        return self.parse_dense_int_or_fp_elements_attr(None)

    def _parse_builtin_opaque_attr(self):
        str_lit_list = self.parse_comma_separated_list(
            self.Delimiter.ANGLE, self.parse_str_literal
        )

        if len(str_lit_list) != 2:
            self.raise_error("Opaque expects 2 string literal parameters!")

        type = NoneAttr()
        if self.parse_optional_punctuation(":") is not None:
            type = self.expect(
                self.parse_optional_type, "opaque attribute must be typed!"
            )

        return OpaqueAttr.from_strings(*str_lit_list, type=type)

    def _parse_builtin_dense_resource_attr(self) -> DenseResourceAttr:
        self.parse_characters("<", " in dense_resource attribute")
        resource_handle = self.parse_identifier(" for resource handle")
        self.parse_characters(">", " in dense_resource attribute")
        self.parse_characters(":", " in dense_resource attribute")
        type = self.parse_type()
        return DenseResourceAttr.from_params(resource_handle, type)

    def _parse_typed_integer(
        self,
        type: IntegerType,
        allow_boolean: bool = True,
        allow_negative: bool = True,
        context_msg: str = "",
    ) -> int:
        """
        Parse an (possible negative) integer. The integer can
        either be decimal or hexadecimal.
        Optionally allow parsing of 'true' or 'false' into 1 and 0.
        """

        pos = self.pos
        res = self.parse_integer(
            allow_boolean=allow_boolean,
            allow_negative=allow_negative,
            context_msg=context_msg,
        )

        try:
            type.verify_value(res)
        except VerifyException as e:
            self.raise_error(str(e), pos, self.pos)

        return res

    def _parse_builtin_densearray_attr(self) -> DenseArrayBase | None:
        self.parse_characters("<", " in dense array")
        pos = self.pos
        element_type = self.parse_attribute()

        if not isinstance(element_type, IntegerType | AnyFloat):
            self.raise_error(
                "dense array element type must be an integer or floating point type",
                pos,
                self.pos,
            )

        # Empty array
        if self.parse_optional_punctuation(">"):
            return DenseArrayBase.from_list(element_type, [])

        self.parse_characters(":", " in dense array")

        if isinstance(element_type, IntegerType):
            values = self.parse_comma_separated_list(
                self.Delimiter.NONE,
                lambda: self._parse_typed_integer(element_type, allow_boolean=True),
            )
            res = DenseArrayBase.create_dense_int(element_type, values)
        else:
            values = self.parse_comma_separated_list(
                self.Delimiter.NONE,
                lambda: self.parse_float(),
            )
            res = DenseArrayBase.create_dense_float(element_type, values)

        self.parse_characters(">", " in dense array")

        return res

    def _parse_builtin_affine_map(self) -> AffineMapAttr:
        self.parse_characters("<", " in affine_map attribute")
        affine_map = self.parse_affine_map()
        self.parse_characters(">", " in affine_map attribute")
        return AffineMapAttr(affine_map)

    def _parse_builtin_affine_set(self) -> AffineSetAttr:
        self.parse_characters("<", " in affine_set attribute")
        affine_set = self.parse_affine_set()
        self.parse_characters(">", " in affine_set attribute")
        return AffineSetAttr(affine_set)

    @dataclass
    class _TensorLiteralElement:
        """
        The representation of a tensor literal element used during parsing.
        It is either an integer, float, or boolean. It also has a check if
        the element has a negative sign (it is already applied to the value).
        This class is used to parse a tensor literal before the tensor literal
        type is known
        """

        is_negative: bool
        value: int | float | bool
        """
        An integer, float, boolean, integer complex, or float complex value.
        The tuple should be of type `_TensorLiteralElement`, but python does
        not allow classes to self-reference.
        """
        span: Span

        def to_int(
            self,
            parser: AttrParser,
            allow_negative: bool = True,
            allow_booleans: bool = True,
        ) -> int:
            """
            Convert the element to an int value, possibly disallowing negative
            values. Raises an error if the type is compatible.
            """
            if self.is_negative and not allow_negative:
                parser.raise_error(
                    "Expected non-negative integer values", at_position=self.span
                )
            if isinstance(self.value, bool) and not allow_booleans:
                parser.raise_error(
                    "Boolean values are only allowed for i1 types",
                    at_position=self.span,
                )
            if not isinstance(self.value, bool | int):
                parser.raise_error("Expected integer value", at_position=self.span)
            return int(self.value)

        def to_float(self, parser: AttrParser) -> float:
            """
            Convert the element to a float value. Raises an error if the type
            is compatible.
            """
            return float(self.value)

        def to_type(self, parser: AttrParser, type: AnyFloat | IntegerType | IndexType):
            if isinstance(type, AnyFloat):
                return self.to_float(parser)

            match type:
                case IntegerType():
                    return self.to_int(
                        parser,
                        type.signedness.data != Signedness.UNSIGNED,
                        type.width.data == 1,
                    )
                case IndexType():
                    return self.to_int(
                        parser, allow_negative=True, allow_booleans=False
                    )

    def _parse_tensor_literal_element(self) -> _TensorLiteralElement:
        """
        Parse a tensor literal element, which can be a boolean, an integer
        literal, or a float literal.
        """
        # boolean case
        if self._current_token.text == "true":
            token = self._consume_token(MLIRTokenKind.BARE_IDENT)
            return self._TensorLiteralElement(False, True, token.span)
        if self._current_token.text == "false":
            token = self._consume_token(MLIRTokenKind.BARE_IDENT)
            return self._TensorLiteralElement(False, False, token.span)

        # checking for negation
        minus_token = self._parse_optional_token(MLIRTokenKind.MINUS)

        # Integer and float case
        if self._current_token.kind == MLIRTokenKind.FLOAT_LIT:
            token = self._consume_token(MLIRTokenKind.FLOAT_LIT)
            value = token.kind.get_float_value(token.span)
        elif self._current_token.kind == MLIRTokenKind.INTEGER_LIT:
            token = self._consume_token(MLIRTokenKind.INTEGER_LIT)
            value = token.kind.get_int_value(token.span)
        else:
            self.raise_error("Expected either a float, integer, or complex literal")

        if minus_token is None:
            is_negative = False
            span = token.span
        else:
            is_negative = True
            value = -value
            span = Span(minus_token.span.start, token.span.end, token.span.input)

        return self._TensorLiteralElement(is_negative, value, span)

    def _parse_tensor_literal(
        self,
    ) -> tuple[list[AttrParser._TensorLiteralElement], list[int]]:
        """
        Parse a tensor literal, and returns its flatten data and its shape.

        For instance, [[0, 1, 2], [3, 4, 5]] will return [0, 1, 2, 3, 4, 5] for
        the data, and [2, 3] for the shape.
        """
        res = self.parse_optional_comma_separated_list(
            self.Delimiter.SQUARE, self._parse_tensor_literal
        )
        if res is not None:
            if len(res) == 0:
                return [], [0]
            sub_literal_shape = res[0][1]
            if any(r[1] != sub_literal_shape for r in res):
                self.raise_error(
                    "Tensor literal has inconsistent ranks between elements"
                )
            shape = [len(res)] + sub_literal_shape
            values = [elem for sub_list in res for elem in sub_list[0]]
            return values, shape
        else:
            element = self._parse_tensor_literal_element()
            return [element], []

    def parse_optional_visibility_keyword(self) -> StringAttr | None:
        """
        Parses the visibility keyword of a symbol if present.
        """
        if self.parse_optional_keyword("public"):
            return StringAttr("public")
        elif self.parse_optional_keyword("nested"):
            return StringAttr("nested")
        elif self.parse_optional_keyword("private"):
            return StringAttr("private")
        else:
            return None

    def parse_visibility_keyword(self) -> StringAttr:
        """
        Parses the visibility keyword of a symbol.
        """
        return self.expect(
            self.parse_optional_visibility_keyword, "expect symbol visibility keyword"
        )

    def parse_optional_symbol_name(self) -> StringAttr | None:
        """
        Parse an @-identifier if present, and return its name (without the '@') in a
        string attribute.
        """
        if (token := self._parse_optional_token(MLIRTokenKind.AT_IDENT)) is None:
            return None

        assert len(token.text) > 1, "token should be at least 2 characters long"

        # In the case where the symbol name is quoted, remove the quotes and escape
        # sequences.
        if token.text[1] == '"':
            literal_span = StringLiteral(
                token.span.start + 1, token.span.end, token.span.input
            )
            return StringAttr(literal_span.string_contents)
        return StringAttr(token.text[1:])

    def parse_symbol_name(self) -> StringAttr:
        """
        Parse an @-identifier and return its name (without the '@') in a string
        attribute.
        """
        return self.expect(self.parse_optional_symbol_name, "expect symbol name")

    def _parse_optional_symref_attr(self) -> SymbolRefAttr | None:
        """
        Parse a symbol reference attribute, if present.
          symbol-attr ::= symbol-ref-id (`::` symbol-ref-id)*
          symbol-ref-id ::= at-ident
        """
        # Parse the root symbol
        sym_root = self.parse_optional_symbol_name()
        if sym_root is None:
            return None

        # Parse nested symbols
        refs: list[StringAttr] = []
        while self._current_token.kind == MLIRTokenKind.COLON:
            # Parse `::`. As in MLIR, this require to backtrack if a single `:` is given.
            pos = self._current_token.span.start
            self._consume_token(MLIRTokenKind.COLON)
            if self._parse_optional_token(MLIRTokenKind.COLON) is None:
                self._resume_from(pos)
                break

            refs.append(self.parse_symbol_name())

        return SymbolRefAttr(sym_root, ArrayAttr(refs))

    def parse_optional_location(self) -> LocationAttr | None:
        """
        Parse a location attribute, if present.
          location ::= `loc` `(` `unknown` `)`
        """
        snapshot = self._current_token.span.start
        if (
            self.parse_optional_characters("loc")
            and self.parse_optional_punctuation("(")
            and self.parse_optional_characters("unknown")
            and self.parse_optional_punctuation(")")
        ):
            return LocationAttr()
        self._resume_from(snapshot)
        return None

    def parse_optional_builtin_int_or_float_attr(
        self,
    ) -> IntegerAttr | FloatAttr | None:
        bool = self.try_parse_builtin_boolean_attr()
        if bool is not None:
            return bool

        is_hexadecimal_token: bool = self._current_token.text[:2] in ["0x", "0X"]

        # Parse the value
        if (value := self.parse_optional_number()) is None:
            return None

        # If no types are given, we take the default ones
        if self._current_token.kind != MLIRTokenKind.COLON:
            if isinstance(value, float):
                return FloatAttr(value, Float64Type())
            return IntegerAttr(value, i64)

        # Otherwise, we parse the attribute type
        type = self._parse_attribute_type()

        if isinstance(type, AnyFloat):
            if is_hexadecimal_token:
                assert isinstance(value, int)
                match type:
                    case Float16Type():
                        return FloatAttr(convert_u16_to_f16(value), type)
                    case Float32Type():
                        return FloatAttr(convert_u32_to_f32(value), type)
                    case Float64Type():
                        return FloatAttr(convert_u64_to_f64(value), type)
                    case _:
                        raise NotImplementedError(
                            f"Cannot parse hexadecimal literal for float type of bit width {type}"
                        )
            return FloatAttr(float(value), type)

        if isinstance(type, IntegerType | IndexType):
            if isinstance(value, float):
                self.raise_error("Floating point value is not valid for integer type.")
            return IntegerAttr(value, type)

        self.raise_error("Invalid type given for integer or float attribute.")

    def try_parse_builtin_boolean_attr(
        self,
    ) -> IntegerAttr[IntegerType | IndexType] | None:
        if (value := self.parse_optional_boolean()) is not None:
            return IntegerAttr(1 if value else 0, IntegerType(1))
        return None

    def _parse_optional_none_type(self) -> NoneType | None:
        """
        Parse a none type, if present
        none-type ::= `none`
        """
        if self.parse_optional_keyword("none"):
            return NoneType()
        return None

    def _parse_optional_string_attr(self) -> StringAttr | None:
        """
        Parse a string attribute, if present.
          string-attr ::= string-literal
        """
        token = self._parse_optional_token(MLIRTokenKind.STRING_LIT)
        return (
            StringAttr(token.kind.get_string_literal_value(token.span))
            if token is not None
            else None
        )

    def _parse_optional_array_attr(self) -> ArrayAttr | None:
        """
        Parse an array attribute, if present, with format:
            array-attr ::= `[` (attribute (`,` attribute)*)? `]`
        """
        attrs = self.parse_optional_comma_separated_list(
            self.Delimiter.SQUARE, self.parse_attribute
        )
        if attrs is None:
            return None
        return ArrayAttr(attrs)

    def parse_function_type(self) -> FunctionType:
        """
        Parse a function type.
            function-type ::= type-list `->` (type | type-list)
            type-list     ::= `(` `)` | `(` type (`,` type)* `)`
        """
        return self.expect(
            self.parse_optional_function_type,
            "function type expected",
        )

    def parse_optional_function_type(self) -> FunctionType | None:
        """
        Parse a function type, if present.
            function-type ::= type-list `->` (type | type-list)
            type-list     ::= `(` `)` | `(` type (`,` type)* `)`
        """
        if self._current_token.kind != MLIRTokenKind.L_PAREN:
            return None

        # Parse the arguments
        args = self.parse_comma_separated_list(self.Delimiter.PAREN, self.parse_type)

        self.parse_punctuation("->")

        # Parse the returns
        returns = self.parse_optional_comma_separated_list(
            self.Delimiter.PAREN, self.parse_type
        )
        if returns is None:
            returns = [self.parse_type()]
        return FunctionType.from_lists(args, returns)

    def parse_paramattr_parameters(
        self, skip_white_space: bool = True
    ) -> list[Attribute]:
        res = self.parse_optional_comma_separated_list(
            self.Delimiter.ANGLE, self.parse_attribute
        )
        if res is None:
            return []
        return res

    def _parse_optional_builtin_dict_attr(self) -> DictionaryAttr | None:
        """
        Parse a dictionary attribute, if present, with format:
        `dictionary-attr ::= `{` ( attribute-entry (`,` attribute-entry)* )? `}`
        `attribute-entry` := (bare-id | string-literal) `=` attribute
        """
        if self._current_token.kind != MLIRTokenKind.L_BRACE:
            return None
        param = DictionaryAttr.parse_parameter(self)
        return DictionaryAttr(param)

    def _parse_builtin_dict_attr(self) -> DictionaryAttr:
        """
        Parse a dictionary attribute with format:
        `dictionary-attr ::= `{` ( attribute-entry (`,` attribute-entry)* )? `}`
        `attribute-entry` := (bare-id | string-literal) `=` attribute
        """
        param = DictionaryAttr.parse_parameter(self)
        return DictionaryAttr(param)

    _builtin_integer_type_regex = re.compile(r"^[su]?i(\d+)$")
    _builtin_float_type_regex = re.compile(r"^f(\d+)$")

    def _parse_optional_integer_or_float_type(self) -> Attribute | None:
        """
        Parse as integer or float type, if present.
          integer-or-float-type ::= index-type | integer-type | float-type
          index-type            ::= `index`
          integer-type          ::= (`i` | `si` | `ui`) decimal-literal
          float-type            ::= `f16` | `f32` | `f64` | `f80` | `f128` | `bf16`
        """
        if self._current_token.kind != MLIRTokenKind.BARE_IDENT:
            return None
        name = self._current_token.text

        # Index type
        if name == "index":
            self._consume_token()
            return IndexType()

        # Integer type
        if (match := self._builtin_integer_type_regex.match(name)) is not None:
            signedness = {
                "s": Signedness.SIGNED,
                "u": Signedness.UNSIGNED,
                "i": Signedness.SIGNLESS,
            }
            self._consume_token()
            return IntegerType(int(match.group(1)), signedness[name[0]])

        # bf16 type
        if name == "bf16":
            self._consume_token()
            return BFloat16Type()

        # Float type
        if (re_match := self._builtin_float_type_regex.match(name)) is not None:
            width = int(re_match.group(1))
            type = {
                16: Float16Type,
                32: Float32Type,
                64: Float64Type,
                80: Float80Type,
                128: Float128Type,
            }.get(width, None)
            if type is None:
                self.raise_error(f"Unsupported floating point width: {width}")
            self._consume_token()
            return type()

        return None

    def _parse_optional_builtin_type(self) -> Attribute | None:
        """
        parse a builtin-type, like i32, index, vector<i32>, none, if present.
        """

        # Check for a function type
        if (function_type := self.parse_optional_function_type()) is not None:
            return function_type

        # Check for an integer or float type
        if (number_type := self._parse_optional_integer_or_float_type()) is not None:
            return number_type

        # check for a none type
        if (none_type := self._parse_optional_none_type()) is not None:
            return none_type

        return self._parse_optional_builtin_parametrized_type()

    def parse_affine_map(self) -> AffineMap:
        affp = affine_parser.AffineParser(self._parser_state)
        return affp.parse_affine_map()

    def parse_affine_set(self) -> AffineSet:
        affp = affine_parser.AffineParser(self._parser_state)
        return affp.parse_affine_set()



from xdsl.dialects import builtin, stencil
from xdsl.dialects.experimental import dmp
from xdsl.ir import Attribute
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa


class DmpSwapShapeInference(RewritePattern):
    """
    Infer the shape of the `dmp.swap` operation.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewrite: PatternRewriter):
        if not op.swapped_values:
            return
        swap_t = op.swapped_values.type
        if not isinstance(swap_t, stencil.TempType):
            return
        if op.input_stencil.type != swap_t:
            op.input_stencil.type = swap_t
            rewrite.handle_operation_modification(op)


class DmpSwapSwapsInference(RewritePattern):
    """
    Infer the exact exchanges this `dmp.swap` needs to perform.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewrite: PatternRewriter):
        core_lb: stencil.IndexAttr | None = None
        core_ub: stencil.IndexAttr | None = None

        if not op.swapped_values:
            return

        for use in op.swapped_values.uses:
            if not isinstance(use.operation, stencil.ApplyOp):
                continue
            assert use.operation.res
            bounds = use.operation.get_bounds()
            if isinstance(bounds, builtin.IntAttr):
                return
            core_lb = bounds.lb
            core_ub = bounds.ub
            break

        # this shouldn't have changed since the op was created!
        temp = op.input_stencil.type
        assert isa(temp, stencil.TempType[Attribute])
        if not isinstance(temp.bounds, stencil.StencilBoundsAttr):
            return
        buff_lb = temp.bounds.lb
        buff_ub = temp.bounds.ub

        if core_lb is None or core_ub is None:
            return

        swaps = builtin.ArrayAttr(
            exchange
            for exchange in op.strategy.halo_exchange_defs(
                dmp.ShapeAttr.from_index_attrs(
                    buff_lb=buff_lb,
                    core_lb=core_lb,
                    buff_ub=buff_ub,
                    core_ub=core_ub,
                )
            )
            # drop empty exchanges
            if exchange.elem_count > 0
        )
        if swaps != op.swaps:
            op.swaps = swaps
            rewrite.handle_operation_modification(op)



"""
This file implements a partial lowering of Toy operations to a combination of
affine loops, memref operations and standard operations. This lowering
expects that all calls have been inlined, and all shapes have been resolved.
"""

from collections.abc import Callable, Sequence
from itertools import product
from typing import TypeAlias, TypeVar, cast

from xdsl.builder import Builder, InsertPoint
from xdsl.context import Context
from xdsl.dialects import affine, arith, func, memref, printf
from xdsl.dialects.builtin import (
    AffineMapAttr,
    Float64Type,
    FloatAttr,
    IndexType,
    IntegerAttr,
    ModuleOp,
    ShapedType,
    f64,
)
from xdsl.ir import Block, Operation, Region, SSAValue
from xdsl.ir.affine import AffineMap
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)

from ..dialects import toy

# region Helpers

MemRefTypeF64: TypeAlias = memref.MemRefType[Float64Type]


def convert_tensor_to_memref(type: toy.TensorTypeF64) -> MemRefTypeF64:
    """
    Convert the given RankedTensorType into the corresponding MemRefType.
    """
    return memref.MemRefType(f64, type.shape)


def insert_alloc_and_dealloc(
    type: MemRefTypeF64, op: Operation, rewriter: PatternRewriter
) -> memref.AllocOp:
    """
    Insert an allocation and deallocation for the given MemRefType.
    """
    block = op.parent

    assert block is not None, "Operation to be allocated must be in a block"
    assert block.last_op is not None

    # Make sure to allocate at the beginning of the block.
    alloc = memref.AllocOp.get(type.element_type, None, type.shape)
    rewriter.insert_op(alloc, InsertPoint.at_start(block))

    # Make sure to deallocate this alloc at the end of the block. This is fine as toy
    # functions have no control flow.
    dealloc = memref.DeallocOp.get(alloc)
    rewriter.insert_op(dealloc, InsertPoint.before(block.last_op))

    return alloc


_ValueRange = Sequence[SSAValue]
_AffineForOpBodyBuilderFn: TypeAlias = Callable[[Builder, SSAValue, _ValueRange], None]


def build_affine_for(
    builder: Builder,
    lb_operands: _ValueRange,
    lb_map: affine.AffineMap,
    ub_operands: _ValueRange,
    ub_map: affine.AffineMap,
    step: int,
    iter_args: _ValueRange,
    body_builder_fn: _AffineForOpBodyBuilderFn,
) -> affine.ForOp:
    """
    `body_builder_fn` is used to build the body of affine.for.
    """

    assert len(lb_operands) == lb_map.num_dims, (
        "lower bound operand count does not match the affine map"
    )
    assert len(ub_operands) == ub_map.num_dims, (
        "upper bound operand count does not match the affine map"
    )
    assert step > 0, "step has to be a positive integer constant"

    # Create a region and a block for the body.
    # The argument of the region is the loop induction variable.
    block_arg_types = (IndexType(), *(arg.type for arg in iter_args))
    block = Block(arg_types=block_arg_types)
    induction_var, *rest = block.args
    region = Region(block)

    op = affine.ForOp.from_region(
        lb_operands,
        ub_operands,
        iter_args,
        tuple(iter_arg.type for iter_arg in iter_args),
        affine.AffineMapAttr(lb_map),
        affine.AffineMapAttr(ub_map),
        region,
        step,
    )
    builder.insert(op)
    body_builder_fn(Builder(InsertPoint.at_end(block)), induction_var, rest)
    return op


def build_affine_for_const(
    builder: Builder,
    lb: int,
    ub: int,
    step: int,
    iter_args: _ValueRange,
    body_builder_fn: _AffineForOpBodyBuilderFn,
) -> affine.ForOp:
    return build_affine_for(
        builder,
        (),
        affine.AffineMap.constant_map(lb),
        (),
        affine.AffineMap.constant_map(ub),
        step,
        iter_args,
        body_builder_fn,
    )


_Bounds: TypeAlias = tuple[int, ...]

LoopIterationFn: TypeAlias = Callable[[Builder, _ValueRange, _ValueRange], SSAValue]
"""
This defines the function type used to process an iteration of a lowered loop. It takes as
input an OpBuilder, an range of memRefOperands corresponding to the operands of the input
operation, and the range of loop induction variables for the iteration. It returns a value
to store at the current index of the iteration.
"""

_BoundT = TypeVar("_BoundT")

_BodyBuilderFn: TypeAlias = Callable[[Builder, _ValueRange], None]
_LoopCreatorFn: TypeAlias = Callable[
    [Builder, _BoundT, _BoundT, int, _AffineForOpBodyBuilderFn],
    affine.ForOp,
]


def build_affine_loop_nest_impl(
    builder: Builder,
    lbs: Sequence[_BoundT],
    ubs: Sequence[_BoundT],
    steps: Sequence[int],
    body_builder_fn: _BodyBuilderFn,
    loop_creator_fn: _LoopCreatorFn[_BoundT],
) -> None:
    """
    Corresponds to affine::buildAffineLoopNestImpl in MLIR
    """

    assert len(lbs) == len(ubs), "Mismatch in number of arguments"
    assert len(lbs) == len(steps), "Mismatch in number of arguments"

    # if there are no loops to be constructed, construct the body anyway
    if not lbs:
        body_builder_fn(builder, ())
        return

    # Create the loops iteratively and store the induction variables.

    ivs: list[SSAValue] = []

    e = len(lbs)
    for i in range(e):
        # Callback for creating the loop body, always creates the terminator.
        def body(nested_builder: Builder, iv: SSAValue, iter_args: _ValueRange):
            nonlocal ivs

            ivs.append(iv)

            # In the innermost loop, call the body builder
            if i == e - 1:
                body_builder_fn(nested_builder, ivs)

            nested_builder.insert(affine.YieldOp.get())

        # Delegate actual loop creation to the callback in order to dispatch
        # between constant- and variable-bound loops.

        loop = loop_creator_fn(builder, lbs[i], ubs[i], steps[i], body)
        builder = Builder(InsertPoint(loop.body.block, loop.body.block.first_op))


def build_affine_loop_from_constants(
    builder: Builder,
    lb: int,
    ub: int,
    step: int,
    body_builder_fn: _AffineForOpBodyBuilderFn,
) -> affine.ForOp:
    """
    Creates an affine loop from the bounds known to be constants.
    """
    return build_affine_for_const(builder, lb, ub, step, (), body_builder_fn)


def build_affine_loop_from_values(
    builder: Builder,
    lb: SSAValue,
    ub: SSAValue,
    step: int,
    body_builder_fn: _AffineForOpBodyBuilderFn,
) -> affine.ForOp:
    """
    Creates an affine loop from the bounds that may or may not be constants.
    """
    lb_const = lb.owner
    ub_const = ub.owner

    if (
        isinstance(lb_const, arith.ConstantOp)
        and isinstance(lb_const_value := lb_const.value, IntegerAttr)
        and isinstance(ub_const, arith.ConstantOp)
        and isinstance(ub_const_value := ub_const.value, IntegerAttr)
    ):
        lb_val = lb_const_value.value.data
        ub_val = ub_const_value.value.data
        return build_affine_loop_from_constants(
            builder, lb_val, ub_val, step, body_builder_fn
        )
    return build_affine_for(
        builder,
        (lb,),
        affine.AffineMap(1, 0, (affine.AffineExpr.dimension(0),)),
        (ub,),
        affine.AffineMap(1, 0, (affine.AffineExpr.dimension(0),)),
        step,
        (),
        body_builder_fn,
    )


def build_affine_loop_nest_const(
    builder: Builder,
    lbs: Sequence[int],
    ubs: Sequence[int],
    steps: Sequence[int],
    body_builder_fn: _BodyBuilderFn,
) -> None:
    build_affine_loop_nest_impl(
        builder, lbs, ubs, steps, body_builder_fn, build_affine_loop_from_constants
    )


def build_affine_loop_nest(
    builder: Builder,
    lbs: Sequence[SSAValue],
    ubs: Sequence[SSAValue],
    steps: Sequence[int],
    body_builder_fn: _BodyBuilderFn,
) -> None:
    build_affine_loop_nest_impl(
        builder, lbs, ubs, steps, body_builder_fn, build_affine_loop_from_values
    )


def lower_op_to_loops(
    op: toy.AddOp | toy.MulOp | toy.TransposeOp,
    operands: _ValueRange,
    rewriter: PatternRewriter,
    process_iteration: LoopIterationFn,
):
    tensor_type = cast(toy.TensorTypeF64, op.res.type)

    # insert an allocation and deallocation for the result of this operation.
    memref_type = convert_tensor_to_memref(tensor_type)
    alloc = insert_alloc_and_dealloc(memref_type, op, rewriter)

    # Create a nest of affine loops, with one loop per dimension of the shape.
    # The buildAffineLoopNest function takes a callback that is used to construct the body
    # of the innermost loop given a builder, a location and a range of loop induction
    # variables.

    rank = tensor_type.get_num_dims()
    lower_bounds = tuple(0 for _ in range(rank))
    steps = tuple(1 for _ in range(rank))

    def impl_loop(nested_builder: Builder, ivs: _ValueRange):
        # Call the processing function with the rewriter, the memref operands, and the
        # loop induction variables. This function will return the value to store at the
        # current index.
        value_to_store = process_iteration(nested_builder, operands, ivs)
        store_op = affine.StoreOp(value_to_store, alloc.memref, ivs)
        nested_builder.insert(store_op)

    builder = Builder(InsertPoint.before(op))
    build_affine_loop_nest_const(
        builder, lower_bounds, tensor_type.get_shape(), steps, impl_loop
    )
    # Replace this operation with the generated alloc.

    op.res.replace_by(alloc.memref)
    rewriter.erase_matched_op()


# endregion Helpers

# region RewritePatterns


class AddOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.AddOp, rewriter: PatternRewriter):
        def body(
            builder: Builder, memref_operands: _ValueRange, loop_ivs: _ValueRange
        ) -> SSAValue:
            # Generate loads for the element of 'lhs' and 'rhs' at the inner loop.
            loaded_lhs = builder.insert(affine.LoadOp(op.lhs, loop_ivs))
            loaded_rhs = builder.insert(affine.LoadOp(op.rhs, loop_ivs))
            new_binop = builder.insert(arith.AddfOp(loaded_lhs, loaded_rhs))
            return new_binop.result

        lower_op_to_loops(op, op.operands, rewriter, body)


class MulOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.MulOp, rewriter: PatternRewriter):
        def body(
            builder: Builder, memref_operands: _ValueRange, loop_ivs: _ValueRange
        ) -> SSAValue:
            # Generate loads for the element of 'lhs' and 'rhs' at the inner loop.
            loaded_lhs = builder.insert(affine.LoadOp(op.lhs, loop_ivs))
            loaded_rhs = builder.insert(affine.LoadOp(op.rhs, loop_ivs))
            new_binop = builder.insert(arith.MulfOp(loaded_lhs, loaded_rhs))
            return new_binop.result

        lower_op_to_loops(op, op.operands, rewriter, body)


class ConstantOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.ConstantOp, rewriter: PatternRewriter):
        constant_value = op.value

        # When lowering the constant operation, we allocate and assign the constant
        # values to a corresponding memref allocation.

        tensor_type = cast(toy.TensorTypeF64, op.res.type)
        memref_type = convert_tensor_to_memref(tensor_type)
        alloc = insert_alloc_and_dealloc(memref_type, op, rewriter)

        value_shape = memref_type.get_shape()

        # Scalar constant values for elements of the tensor
        constants: list[arith.ConstantOp] = [
            arith.ConstantOp(FloatAttr(i, f64)) for i in constant_value.get_values()
        ]

        # n-d indices of elements
        _indices = product(*(range(d) for d in value_shape))

        # For each n-d index into the tensor, store the corresponding scalar
        stores = [
            affine.StoreOp(
                constants[offset].result,
                alloc.memref,
                (),
                map=affine.AffineMapAttr(affine.AffineMap.point_map(*index)),
            )
            for offset, index in enumerate(_indices)
        ]

        # Insert constants used before the alloc, not before matched operation
        rewriter.insert_op(constants, InsertPoint.before(alloc))

        # Replace the constant by the stores, and its result by the allocated value
        rewriter.replace_matched_op(stores, (alloc.memref,))


class FuncOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.FuncOp, rewriter: PatternRewriter):
        name = op.sym_name.data

        # We only lower the main function as we expect that all other functions
        # have been inlined.
        assert name == "main", "Only support lowering main function for now"

        # Verify that the given main has no inputs and results.
        if op.function_type.inputs or op.function_type.outputs:
            raise ValueError("expected 'main' to have 0 inputs and 0 results")

        # Create a new non-toy function, with the same region.
        region = op.regions[0]

        new_op = func.FuncOp(
            name, op.function_type, rewriter.move_region_contents_to_new_regions(region)
        )

        rewriter.replace_matched_op(new_op)


class PrintOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.PrintOp, rewriter: PatternRewriter):
        assert isinstance(shaped_type := op.input.type, ShapedType)
        shape = shaped_type.get_shape()

        format_str = "{}"

        for dim in reversed(shape):
            format_str = "[" + ", ".join([format_str] * dim) + "]"

        new_vals: list[SSAValue] = []

        for indices in product(*(range(dim) for dim in shape)):
            rewriter.insert_op_before_matched_op(
                load := affine.LoadOp(
                    op.input,
                    (),
                    AffineMapAttr(AffineMap.from_callable(lambda: indices)),
                )
            )
            new_vals.append(load.result)

        rewriter.replace_matched_op(printf.PrintFormatOp(format_str, *new_vals))


class ReturnOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.ReturnOp, rewriter: PatternRewriter):
        assert op.input is None, (
            "During this lowering, we expect that all function calls have been inlined."
        )

        rewriter.replace_matched_op(func.ReturnOp())


class TransposeOpLowering(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.TransposeOp, rewriter: PatternRewriter):
        def body(
            builder: Builder, mem_ref_operands: _ValueRange, loop_ivs: _ValueRange
        ) -> SSAValue:
            # Transpose the elements by generating a load from the reverse indices.
            load_op = affine.LoadOp(op.arg, tuple(reversed(loop_ivs)))
            builder.insert(load_op)
            return load_op.result

        lower_op_to_loops(op, op.operands, rewriter, body)


# endregion RewritePatterns


class LowerToAffinePass(ModulePass):
    """
    A pass for lowering operations in the Toy dialect to Builtin.
    """

    name = "toy-to-builtin"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    AddOpLowering(),
                    ConstantOpLowering(),
                    FuncOpLowering(),
                    MulOpLowering(),
                    PrintOpLowering(),
                    ReturnOpLowering(),
                    TransposeOpLowering(),
                ]
            )
        ).rewrite_module(op)



from collections.abc import Callable

from xdsl.interactive.passes import (
    AvailablePass,
    apply_passes_to_module,
    get_condensed_pass_list,
    get_new_registered_context,
)
from xdsl.interactive.rewrites import get_all_possible_rewrites
from xdsl.ir import Dialect
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import RewritePattern


def get_available_pass_list(
    all_dialects: tuple[tuple[str, Callable[[], Dialect]], ...],
    all_passes: tuple[tuple[str, type[ModulePass]], ...],
    input_text: str,
    pass_pipeline: tuple[ModulePass, ...],
    condense_mode: bool,
    rewrite_by_names_dict: dict[str, dict[str, RewritePattern]],
) -> tuple[AvailablePass, ...]:
    """
    This function returns the available pass list file based on an input text string, pass_pipeline and condense_mode.
    """
    ctx = get_new_registered_context(all_dialects)
    parser = Parser(ctx, input_text)
    current_module = parser.parse_module()

    current_module = apply_passes_to_module(current_module, ctx, pass_pipeline)

    # get all individual rewrites
    individual_rewrites = get_all_possible_rewrites(
        current_module,
        rewrite_by_names_dict,
    )
    # merge rewrite passes with "other" pass list
    if condense_mode:
        pass_list = get_condensed_pass_list(current_module, all_passes)
    else:
        pass_list = tuple(AvailablePass(p.name, p) for _, p in all_passes)
    return pass_list + tuple(individual_rewrites)



import re

import pytest

from xdsl.backend.riscv.riscv_register_queue import RiscvRegisterQueue
from xdsl.dialects import riscv
from xdsl.dialects.builtin import IntAttr


def test_default_reserved_registers():
    register_queue = RiscvRegisterQueue.default()

    for reg in (
        riscv.Registers.ZERO,
        riscv.Registers.SP,
        riscv.Registers.GP,
        riscv.Registers.TP,
        riscv.Registers.FP,
        riscv.Registers.S0,
    ):
        available_before = register_queue.available_int_registers.copy()
        register_queue.push(reg)
        assert available_before == register_queue.available_int_registers


def test_push_j_register():
    register_queue = RiscvRegisterQueue()

    j0 = riscv.IntRegisterType.infinite_register(0)
    register_queue.push(j0)
    assert register_queue.pop(riscv.IntRegisterType) == j0

    fj0 = riscv.FloatRegisterType.infinite_register(0)
    register_queue.push(fj0)
    assert register_queue.pop(riscv.FloatRegisterType) == fj0


def test_push_register():
    register_queue = RiscvRegisterQueue()

    register_queue.push(riscv.Registers.A0)
    assert register_queue.pop(riscv.IntRegisterType) == riscv.Registers.A0

    register_queue.push(riscv.Registers.FA0)
    assert register_queue.pop(riscv.FloatRegisterType) == riscv.Registers.FA0


def test_reserve_register():
    register_queue = RiscvRegisterQueue()

    j0 = riscv.IntRegisterType.infinite_register(0)
    assert isinstance(j0.index, IntAttr)

    register_queue.reserve_register(j0)
    assert register_queue.reserved_int_registers[j0.index.data] == 1

    register_queue.reserve_register(j0)
    assert register_queue.reserved_int_registers[j0.index.data] == 2

    register_queue.unreserve_register(j0)
    assert register_queue.reserved_int_registers[j0.index.data] == 1

    register_queue.unreserve_register(j0)
    assert j0 not in register_queue.reserved_int_registers
    assert j0 not in register_queue.available_int_registers

    # Check assertion error when reserving an available register
    reg = register_queue.pop(riscv.IntRegisterType)
    register_queue.push(reg)
    register_queue.reserve_register(reg)
    with pytest.raises(
        AssertionError,
        match=re.escape(
            f"Cannot pop a reserved register ({reg.register_name.data}), it must have been reserved while available."
        ),
    ):
        register_queue.pop(riscv.IntRegisterType)


def test_limit():
    register_queue = RiscvRegisterQueue.default()
    register_queue.limit_registers(1)

    assert not register_queue.pop(riscv.IntRegisterType).register_name.data.startswith(
        "j"
    )
    assert register_queue.pop(riscv.IntRegisterType).register_name.data.startswith("j")

    assert not register_queue.pop(
        riscv.FloatRegisterType
    ).register_name.data.startswith("fj")
    assert register_queue.pop(riscv.FloatRegisterType).register_name.data.startswith(
        "fj"
    )



from __future__ import annotations

import math
from collections.abc import Mapping, Sequence
from typing import Any, cast

from typing_extensions import Self

from xdsl.dialects import memref
from xdsl.dialects.builtin import (
    Annotated,
    AnySignlessIntegerOrIndexType,
    ArrayAttr,
    ContainerType,
    DenseArrayBase,
    IndexType,
    IntegerAttr,
    IntegerType,
    TensorType,
    UnrankedTensorType,
    i64,
)
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AttrSizedOperandSegments,
    IRDLOperation,
    Operand,
    base,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
    traits_def,
    var_operand_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.traits import NoMemoryEffect
from xdsl.utils.exceptions import VerifyException


@irdl_op_definition
class CastOp(IRDLOperation):
    """
    Convert a tensor from one type to an equivalent type without changing any data elements.
    The source and destination types must both be tensor types with the same element type.
    If both are ranked, then the rank should be the same and static dimensions should match.
    The operation is invalid if converting to a mismatching constant dimension.
    """

    name = "tensor.cast"

    source = operand_def(
        base(TensorType[Attribute]) | base(UnrankedTensorType[Attribute])
    )
    dest = result_def(base(TensorType[Attribute]) | base(UnrankedTensorType[Attribute]))

    assembly_format = "$source attr-dict `:` type($source) `to` type($dest)"

    traits = traits_def(NoMemoryEffect())

    def __init__(self, source: SSAValue | Operation, dest: TensorType[Attribute]):
        super().__init__(operands=(source,), result_types=(dest,))

    def verify_(self):
        source_type = self.source.type
        dest_type = self.dest.type

        if isinstance(source_type, TensorType) and isinstance(dest_type, TensorType):
            # rank should be the same + constant shapes equal
            if len(source_type.get_shape()) != (len(dest_type.get_shape())):
                raise VerifyException("source and destination rank should be the same")
            for a, b in zip(source_type.get_shape(), dest_type.get_shape()):
                if a >= 0 and b >= 0 and a != b:
                    raise VerifyException(
                        "source and destination constant dimensions should match"
                    )


@irdl_op_definition
class DimOp(IRDLOperation):
    """
    The tensor.dim operation takes a tensor and a dimension operand of type index.
    It returns the size of the requested dimension of the given tensor.
    If the dimension index is out of bounds, the behavior is undefined
    """

    name = "tensor.dim"

    source = operand_def(
        base(TensorType[Attribute]) | base(UnrankedTensorType[Attribute])
    )
    index = operand_def(IndexType)
    result = result_def(IndexType)

    traits = traits_def(NoMemoryEffect())

    def __init__(
        self,
        source: SSAValue | Operation,
        index: SSAValue | Operation,
        attributes: Mapping[str, Attribute] | None = None,
    ):
        super().__init__(
            operands=(source, index), result_types=(IndexType(),), attributes=attributes
        )

    def print(self, printer: Printer):
        printer.print_op_attributes(self.attributes)
        printer.print_string(" ")
        printer.print_ssa_value(self.source)
        printer.print_string(", ")
        printer.print_ssa_value(self.index)
        printer.print_string(" : ")
        printer.print_attribute(self.source.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        attributes = parser.parse_optional_attr_dict()
        source = parser.parse_operand()
        parser.parse_punctuation(",")
        index = parser.parse_operand()
        parser.parse_punctuation(":")
        parser.parse_type()
        return cls(source, index, attributes)

    def verify_(self):
        if isinstance((source_type := self.source.type), TensorType):
            if not len(source_type.get_shape()):
                raise VerifyException("cannot get dim of 0-rank tensor")


@irdl_op_definition
class EmptyOp(IRDLOperation):
    name = "tensor.empty"

    dynamic_sizes = var_operand_def(IndexType)

    tensor = result_def(TensorType[Attribute])

    traits = traits_def(NoMemoryEffect())

    def __init__(self, dynamic_sizes: Sequence[SSAValue], tensor_type: Attribute):
        super().__init__(
            operands=(dynamic_sizes,),
            result_types=(tensor_type,),
        )

    def print(self, printer: Printer):
        if self.dynamic_sizes:
            printer.print_string("(")
            printer.print_list(self.dynamic_sizes, printer.print_ssa_value)
            printer.print_string(")")
        else:
            printer.print_string("(")
            printer.print_string(")")

        printer.print_string(" : ")
        printer.print_attribute(self.tensor.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        pos = parser.pos
        parser.parse_punctuation("(")
        if parser.parse_optional_punctuation(")"):
            dynamic_sizes = ()
        else:
            unresolved_dynamic_sizes = parser.parse_comma_separated_list(
                Parser.Delimiter.NONE, parser.parse_unresolved_operand
            )
            unresolved_types = (IndexType(),) * len(unresolved_dynamic_sizes)
            parser.parse_punctuation(")")
            dynamic_sizes = parser.resolve_operands(
                unresolved_dynamic_sizes, unresolved_types, pos
            )
        parser.parse_punctuation(":")
        result_type = parser.parse_attribute()

        empty = cls(dynamic_sizes, result_type)

        return empty


ReassociationAttr = ArrayAttr[
    ArrayAttr[IntegerAttr[Annotated[IntegerType, IntegerType(64)]]]
]


@irdl_op_definition
class CollapseShapeOp(IRDLOperation):
    name = "tensor.collapse_shape"

    src = operand_def(TensorType[Attribute])
    result = result_def(TensorType[Attribute])
    reassociation = prop_def(ReassociationAttr)
    assembly_format = (
        "$src $reassociation attr-dict `:` type($src) `into` type($result)"
    )

    traits = traits_def(NoMemoryEffect())


@irdl_op_definition
class ReshapeOp(IRDLOperation):
    name = "tensor.reshape"

    source = operand_def(TensorType[Attribute])
    shape = operand_def(TensorType[AnySignlessIntegerOrIndexType])
    result = result_def(TensorType[Attribute])

    traits = traits_def(NoMemoryEffect())

    def __init__(self, source: SSAValue, shape: SSAValue, result_type: Attribute):
        super().__init__(
            operands=(
                source,
                shape,
            ),
            result_types=(result_type,),
        )

    def print(self, printer: Printer):
        printer.print_string(" ")
        printer.print_ssa_value(self.source)
        printer.print_string("(")
        printer.print_ssa_value(self.shape)
        printer.print_string(")")
        printer.print_string(" : ")
        printer.print_string("(")
        printer.print_attribute(self.source.type)
        printer.print_string(", ")
        printer.print_attribute(self.shape.type)
        printer.print_string(")")
        printer.print_string(" -> ")
        printer.print_attribute(self.result.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        attrs = parser.parse_optional_attr_dict()
        source = parser.parse_operand()
        parser.parse_punctuation("(")
        shape = parser.parse_operand()
        parser.parse_punctuation(")")
        parser.parse_punctuation(":")
        parser.parse_punctuation("(")
        parser.parse_comma_separated_list(Parser.Delimiter.NONE, parser.parse_type)
        parser.parse_punctuation(")")
        parser.parse_optional_punctuation("->")
        result_type = parser.parse_attribute()

        reshape = cls(
            source,
            shape,
            result_type,
        )
        reshape.attributes |= attrs
        return reshape

    def verify_(self) -> None:
        if (
            not isinstance(source_type := self.source.type, TensorType)
            or not isinstance(shape_type := self.shape.type, TensorType)
            or not isinstance(res_type := self.result.type, TensorType)
        ):
            raise ValueError(
                "tensor elementwise operation operands and result must be of type TensorType"
            )

        source_type = cast(TensorType[Attribute], source_type)
        shape_type = cast(TensorType[Attribute], shape_type)
        res_type = cast(TensorType[Attribute], res_type)

        if source_type.element_type != res_type.element_type:
            raise VerifyException(
                "element types of source and result tensor types should be the same"
            )

        source_type = source_type.get_shape()
        shape_type = shape_type.get_shape()
        res_type = res_type.get_shape()

        if len(shape_type) != 1:
            raise VerifyException("shape tensor must have a rank one")

        # concerns the case of static reshaping
        if math.prod(source_type) != math.prod(res_type):
            raise VerifyException(
                "source and result tensor should have the same number of elements"
            )

        shape_size = shape_type[0]
        if shape_size != len(res_type):
            raise VerifyException(
                "length of shape operand differs from the result's tensor rank"
            )


@irdl_op_definition
class ExtractSliceOp(IRDLOperation):
    name = "tensor.extract_slice"

    source = operand_def(TensorType)
    offsets = var_operand_def(IndexType)
    sizes = var_operand_def(IndexType)
    strides = var_operand_def(IndexType)
    static_offsets = prop_def(DenseArrayBase)
    static_sizes = prop_def(DenseArrayBase)
    static_strides = prop_def(DenseArrayBase)
    result = result_def(TensorType)

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def from_static_parameters(
        source: SSAValue | Operation,
        offsets: Sequence[int],
        sizes: Sequence[int],
        strides: Sequence[int] | None = None,
        reduce_rank: bool = False,
    ) -> ExtractSliceOp:
        if strides is None:
            strides = [1] * len(offsets)
        source_v = SSAValue.get(source)
        source_t = source_v.type
        if not isinstance(source_t, ContainerType):
            raise ValueError(f"Expected ContainerType, got {source_t}")

        if reduce_rank:
            result_sizes = list(s for s in sizes if s != 1)
        else:
            result_sizes = list(sizes)

        return_type = TensorType[Any](source_t.get_element_type(), result_sizes)

        return ExtractSliceOp.build(
            operands=[source, [], [], []],
            result_types=[return_type],
            properties={
                "static_offsets": DenseArrayBase.from_list(i64, offsets),
                "static_sizes": DenseArrayBase.from_list(i64, result_sizes),
                "static_strides": DenseArrayBase.from_list(i64, strides),
            },
        )


@irdl_op_definition
class InsertSliceOp(IRDLOperation):
    name = "tensor.insert_slice"

    source = operand_def(TensorType)
    dest = operand_def(TensorType)
    offsets = var_operand_def(IndexType)
    sizes = var_operand_def(IndexType)
    strides = var_operand_def(IndexType)
    static_offsets = prop_def(DenseArrayBase)
    static_sizes = prop_def(DenseArrayBase)
    static_strides = prop_def(DenseArrayBase)
    result = result_def(TensorType)

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def get(
        source: Operand,
        dest: Operand,
        static_sizes: Sequence[int],
        static_offsets: Sequence[int] | None = None,
        static_strides: Sequence[int] | None = None,
        offsets: Sequence[Operand] | None = None,
        sizes: Sequence[Operand] | None = None,
        strides: Sequence[Operand] | None = None,
        result_type: Attribute | None = None,
    ) -> InsertSliceOp:
        dims = len(static_sizes)
        offsets = [] if offsets is None else offsets
        sizes = [] if sizes is None else sizes
        strides = [] if strides is None else strides
        if not static_offsets:
            static_offsets = [memref.SubviewOp.DYNAMIC_INDEX] * len(offsets) + (
                [0] * (dims - len(offsets))
            )
        if not static_strides:
            static_strides = [memref.SubviewOp.DYNAMIC_INDEX] * len(strides) + (
                [1] * (dims - len(strides))
            )
        return InsertSliceOp.build(
            operands=[
                source,
                dest,
                offsets,
                sizes,
                strides,
            ],
            properties={
                "static_offsets": DenseArrayBase.from_list(
                    i64,
                    static_offsets,
                ),
                "static_sizes": DenseArrayBase.from_list(
                    i64,
                    static_sizes,
                ),
                "static_strides": DenseArrayBase.from_list(
                    i64,
                    static_strides,
                ),
            },
            result_types=[result_type if result_type else dest.type],
        )

    @staticmethod
    def from_static_parameters(
        source: SSAValue | Operation,
        dest: SSAValue | Operation,
        offsets: Sequence[int],
        sizes: Sequence[int],
        strides: Sequence[int] | None = None,
    ) -> InsertSliceOp:
        source = SSAValue.get(source)
        dest = SSAValue.get(dest)

        if strides is None:
            strides = [1] * len(sizes)

        return InsertSliceOp.build(
            operands=[source, dest, [], [], []],
            result_types=[dest.type],
            properties={
                "static_offsets": DenseArrayBase.from_list(i64, offsets),
                "static_sizes": DenseArrayBase.from_list(i64, sizes),
                "static_strides": DenseArrayBase.from_list(i64, strides),
            },
        )


@irdl_op_definition
class ExtractOp(IRDLOperation):
    name = "tensor.extract"

    tensor = operand_def(TensorType)
    indices = var_operand_def(IndexType)
    result = result_def(Attribute)
    # assembly_format = "$tensor `[` $indices `]` attr-dict `:` type($tensor)"

    def __init__(
        self,
        tensor: SSAValue,
        indices: Sequence[SSAValue] | SSAValue,
        result_type: Attribute,
    ):
        if isinstance(indices, SSAValue):
            indices = [indices]
        return super().__init__(operands=[tensor, indices], result_types=[result_type])

    def print(self, printer: Printer):
        printer.print_string(" ")
        printer.print_ssa_value(self.tensor)
        printer.print_string("[")
        printer.print_list(self.indices, printer.print_ssa_value)
        printer.print_string("]")
        printer.print_string(" : ")
        printer.print_attribute(self.tensor.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        tensor = parser.parse_operand()
        indices = parser.parse_comma_separated_list(
            delimiter=parser.Delimiter.SQUARE, parse=parser.parse_operand
        )
        parser.parse_punctuation(":")
        source_tensor_type = parser.parse_type()
        tensor_type = cast(TensorType[Attribute], source_tensor_type)
        return cls(tensor, indices, tensor_type.get_element_type())


@irdl_op_definition
class InsertOp(IRDLOperation):
    name = "tensor.insert"

    scalar = operand_def(Attribute)
    dest = operand_def(TensorType)
    indices = var_operand_def(IndexType)
    result = result_def(TensorType)
    # assembly_format = "$scalar `into` $dest `[` $indices `]` attr-dict `:` type($dest)"

    def __init__(
        self,
        scalar: SSAValue,
        dest: SSAValue,
        indices: Sequence[SSAValue] | SSAValue,
    ):
        if isinstance(indices, SSAValue):
            indices = [indices]
        super().__init__(operands=(scalar, dest, indices), result_types=(dest.type,))

    def print(self, printer: Printer):
        printer.print_string(" ")
        printer.print_ssa_value(self.scalar)
        printer.print_string(" into ")
        printer.print_ssa_value(self.dest)
        printer.print_string("[")
        printer.print_list(self.indices, printer.print_ssa_value)
        printer.print_string("]")
        printer.print_string(" : ")
        printer.print_attribute(self.dest.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        scalar = parser.parse_operand()
        parser.parse_characters("into")
        dest = parser.parse_operand()
        indices = parser.parse_comma_separated_list(
            delimiter=parser.Delimiter.SQUARE, parse=parser.parse_operand
        )
        parser.parse_punctuation(":")
        parser.parse_type()
        return cls(scalar, dest, indices)


Tensor = Dialect(
    "tensor",
    [
        CastOp,
        DimOp,
        EmptyOp,
        ExtractSliceOp,
        InsertSliceOp,
        ReshapeOp,
        CollapseShapeOp,
        ExtractOp,
        InsertOp,
    ],
    [],
)



from collections import Counter
from collections.abc import Sequence

import marimo as mo

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass, PipelinePass


def asm_html(asm: str) -> mo.Html:
    """
    Returns a Marimo-optimised representation of the assembly code passed in.
    """
    return mo.ui.code_editor(asm, language="python", disabled=True)


def module_html(module: ModuleOp) -> mo.Html:
    """
    Returns a Marimo-optimised representation of the module passed in.
    """
    return mo.ui.code_editor(str(module), language="javascript", disabled=True)


def _spec_str(p: ModulePass) -> str:
    """
    A string representation of the pass passed in, to display to the user.
    """
    if isinstance(p, PipelinePass):
        return ",".join(str(c.pipeline_pass_spec()) for c in p.passes)
    else:
        return str(p.pipeline_pass_spec())


def pipeline_html(
    ctx: Context, module: ModuleOp, passes: Sequence[tuple[mo.Html, ModulePass]]
) -> tuple[Context, ModuleOp, mo.Html]:
    """
    Returns a tuple of the resulting context and module after applying the passes, and
    the Marimo-optimised representation of the modules throughout compilation.

    The input is the input Context, a sequence of tuples of
    (pass description, module pass).

    Marimo's reactive mechanism relies on a graph of values defined in one cell and used
    in another, and cannot detect mutation by reference, hence the new values instead of
    the usual mutation.
    """
    res = module.clone()
    ctx = ctx.clone()
    d: list[mo.Html] = []
    total_key_count = Counter(_spec_str(p) for _, p in passes)
    d_key_count = Counter[str]()
    for text, p in passes:
        p.apply(ctx, res)
        spec = _spec_str(p)
        d_key_count[spec] += 1
        if total_key_count[spec] != 1:
            header = f"{spec} ({d_key_count[spec]})"
        else:
            header = spec
        html_res = module_html(res)
        d.append(
            mo.vstack(
                (
                    header,
                    text,
                    html_res,
                )
            )
        )
    return (ctx, res, mo.carousel(d))



from pathlib import Path

import pytest

from xdsl.utils.exceptions import ParseError

from ..frontend.location import Location
from ..frontend.parser import ToyParser
from ..frontend.toy_ast import (
    BinaryExprAST,
    CallExprAST,
    FunctionAST,
    LiteralExprAST,
    ModuleAST,
    NumberExprAST,
    PrintExprAST,
    PrototypeAST,
    ReturnExprAST,
    VarDeclExprAST,
    VariableExprAST,
    VarType,
)


def test_parse_ast():
    ast_toy = Path("docs/Toy/examples/ast.toy")

    with open(ast_toy) as f:
        parser = ToyParser(ast_toy, f.read())

    parsed_module_ast = parser.parseModule()

    def loc(line: int, col: int) -> Location:
        return Location(str(ast_toy), line, col)

    module_ast = ModuleAST(
        (
            FunctionAST(
                loc(4, 1),
                PrototypeAST(
                    loc(4, 1),
                    "multiply_transpose",
                    [
                        VariableExprAST(loc(4, 24), "a"),
                        VariableExprAST(loc(4, 27), "b"),
                    ],
                ),
                (
                    ReturnExprAST(
                        loc(5, 3),
                        BinaryExprAST(
                            loc(5, 25),
                            "*",
                            CallExprAST(
                                loc(5, 10),
                                "transpose",
                                [VariableExprAST(loc(5, 20), "a")],
                            ),
                            CallExprAST(
                                loc(5, 25),
                                "transpose",
                                [VariableExprAST(loc(5, 35), "b")],
                            ),
                        ),
                    ),
                ),
            ),
            FunctionAST(
                loc(8, 1),
                PrototypeAST(loc(8, 1), "main", []),
                (
                    VarDeclExprAST(
                        loc(11, 3),
                        "a",
                        VarType([]),
                        LiteralExprAST(
                            loc(11, 11),
                            [
                                LiteralExprAST(
                                    loc(11, 12),
                                    [
                                        NumberExprAST(loc(11, 13), 1.0),
                                        NumberExprAST(loc(11, 16), 2.0),
                                        NumberExprAST(loc(11, 19), 3.0),
                                    ],
                                    [3],
                                ),
                                LiteralExprAST(
                                    loc(11, 23),
                                    [
                                        NumberExprAST(loc(11, 24), 4.0),
                                        NumberExprAST(loc(11, 27), 5.0),
                                        NumberExprAST(loc(11, 30), 6.0),
                                    ],
                                    [3],
                                ),
                            ],
                            [2, 3],
                        ),
                    ),
                    VarDeclExprAST(
                        loc(15, 3),
                        "b",
                        VarType([2, 3]),
                        LiteralExprAST(
                            loc(15, 17),
                            [
                                NumberExprAST(loc(15, 18), 1.0),
                                NumberExprAST(loc(15, 21), 2.0),
                                NumberExprAST(loc(15, 24), 3.0),
                                NumberExprAST(loc(15, 27), 4.0),
                                NumberExprAST(loc(15, 30), 5.0),
                                NumberExprAST(loc(15, 33), 6.0),
                            ],
                            [6],
                        ),
                    ),
                    VarDeclExprAST(
                        loc(19, 3),
                        "c",
                        VarType([]),
                        CallExprAST(
                            loc(19, 11),
                            "multiply_transpose",
                            [
                                VariableExprAST(loc(19, 30), "a"),
                                VariableExprAST(loc(19, 33), "b"),
                            ],
                        ),
                    ),
                    VarDeclExprAST(
                        loc(22, 3),
                        "d",
                        VarType([]),
                        CallExprAST(
                            loc(22, 11),
                            "multiply_transpose",
                            [
                                VariableExprAST(loc(22, 30), "b"),
                                VariableExprAST(loc(22, 33), "a"),
                            ],
                        ),
                    ),
                    VarDeclExprAST(
                        loc(25, 3),
                        "e",
                        VarType([]),
                        CallExprAST(
                            loc(25, 11),
                            "multiply_transpose",
                            [
                                VariableExprAST(loc(25, 30), "b"),
                                VariableExprAST(loc(25, 33), "c"),
                            ],
                        ),
                    ),
                    VarDeclExprAST(
                        loc(28, 3),
                        "f",
                        VarType([]),
                        CallExprAST(
                            loc(28, 11),
                            "multiply_transpose",
                            [
                                CallExprAST(
                                    loc(28, 30),
                                    "transpose",
                                    [VariableExprAST(loc(28, 40), "a")],
                                ),
                                VariableExprAST(loc(28, 44), "c"),
                            ],
                        ),
                    ),
                ),
            ),
        )
    )

    assert parsed_module_ast == module_ast


def test_parse_error():
    program = "def("
    parser = ToyParser(Path(), program)
    with pytest.raises(ParseError, match="Expected expression"):
        parser.parseIdentifierExpr()


def test_parse_scalar():
    ast_toy = Path("docs/Toy/examples/scalar.toy")

    with open(ast_toy) as f:
        parser = ToyParser(ast_toy, f.read())

    parsed_module_ast = parser.parseModule()

    def loc(line: int, col: int) -> Location:
        return Location(str(ast_toy), line, col)

    module_ast = ModuleAST(
        (
            FunctionAST(
                loc(3, 1),
                PrototypeAST(loc(3, 1), "main", []),
                (
                    VarDeclExprAST(
                        loc(4, 3),
                        "a",
                        VarType([2, 2]),
                        NumberExprAST(loc(4, 17), 5.5),
                    ),
                    PrintExprAST(
                        loc(5, 3),
                        VariableExprAST(loc(5, 9), "a"),
                    ),
                ),
            ),
        )
    )

    assert parsed_module_ast.dump() == module_ast.dump()

    assert parsed_module_ast == module_ast



import pytest

from xdsl.dialects.builtin import IntegerType, i32
from xdsl.dialects.comb import (
    AddOp,
    AndOp,
    ConcatOp,
    ICmpOp,
    MulOp,
    OrOp,
    VariadicCombOperation,
    XorOp,
)
from xdsl.dialects.test import TestOp, TestType
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import TestSSAValue


def test_icmp_incorrect_comparison():
    a = TestSSAValue(i32)
    b = TestSSAValue(i32)

    with pytest.raises(VerifyException) as e:
        # 'slet' is an invalid comparison operation
        _icmp_op = ICmpOp(a, b, "slet")
    assert e.value.args[0] == "Unknown comparison mnemonic: slet"


def test_comb_concat_builder():
    a = TestSSAValue(IntegerType(5))
    b = TestSSAValue(IntegerType(3))
    c = TestSSAValue(IntegerType(1))
    foo = TestSSAValue(TestType("foo"))

    concat = ConcatOp.from_int_values([a, b, c])
    assert concat is not None
    assert isinstance(concat.result.type, IntegerType)
    assert concat.result.type.width.data == 9

    bad_concat = ConcatOp.from_int_values([a, foo])
    assert bad_concat is None


def test_comb_concat_verifier():
    a = TestOp(result_types=[IntegerType(5)])
    b = TestOp(result_types=[IntegerType(3)])
    c = TestOp(result_types=[IntegerType(1)])

    ConcatOp([a, b, c], IntegerType(9)).verify()

    with pytest.raises(VerifyException):
        ConcatOp([a, b, c], IntegerType(2)).verify()


@pytest.mark.parametrize(
    "ctor",
    [
        AddOp,
        MulOp,
        AndOp,
        OrOp,
        XorOp,
    ],
)
def test_comb_variadic_builder_verifier(ctor: type[VariadicCombOperation]):
    a = TestSSAValue(IntegerType(5))
    b = TestSSAValue(IntegerType(6))

    ctor([a]).verify()
    ctor([a, a]).verify()
    ctor([a, a, a]).verify()
    ctor([a, a], a.type).verify()

    with pytest.raises(ValueError, match="cannot infer type"):
        ctor([]).verify()

    with pytest.raises(VerifyException, match="op expected 1 or more operands"):
        ctor([], a.type).verify()

    with pytest.raises(VerifyException):
        ctor([a, b]).verify()

    with pytest.raises(VerifyException):
        ctor([a, a], b.type).verify()



from __future__ import annotations

from typing import Annotated, ClassVar, Generic, TypeVar

import pytest

from xdsl.context import Context
from xdsl.dialects.builtin import (
    BoolAttr,
    DenseArrayBase,
    IndexType,
    IntAttr,
    IntegerType,
    StringAttr,
    i32,
    i64,
)
from xdsl.dialects.test import TestType
from xdsl.ir import Attribute, Block, Region
from xdsl.irdl import (
    AnyAttr,
    AnyOf,
    AttributeDef,
    AttrSizedOperandSegments,
    AttrSizedRegionSegments,
    AttrSizedResultSegments,
    BaseAttr,
    ConstraintVar,
    EqAttrConstraint,
    IRDLOperation,
    OpDef,
    OperandDef,
    PropertyDef,
    RangeOf,
    RangeVarConstraint,
    RegionDef,
    ResultDef,
    VarConstraint,
    attr_def,
    base,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    opt_operand_def,
    opt_prop_def,
    opt_region_def,
    opt_result_def,
    prop_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
    var_region_def,
    var_result_def,
)
from xdsl.parser import Parser
from xdsl.traits import NoTerminator
from xdsl.utils.exceptions import (
    DiagnosticException,
    PyRDLOpDefinitionError,
    VerifyException,
)
from xdsl.utils.test_value import TestSSAValue

################################################################################
#                              IRDL definition                                 #
################################################################################


@irdl_op_definition
class OpDefTestOp(IRDLOperation):
    name = "test.op_def_test"

    irdl_options = [AttrSizedOperandSegments()]

    operand = operand_def()
    result = result_def()
    prop = prop_def(Attribute)
    attr = attr_def(Attribute)
    region = region_def()

    # Check that we can define methods in operation definitions
    def test(self):
        pass


def test_get_definition():
    """Test retrieval of an IRDL definition from an operation"""
    assert OpDefTestOp.get_irdl_definition() == OpDef(
        "test.op_def_test",
        operands=[("operand", OperandDef(AnyAttr()))],
        results=[("result", ResultDef(AnyAttr()))],
        attributes={
            "attr": AttributeDef(AnyAttr()),
            "operandSegmentSizes": AttributeDef(BaseAttr(DenseArrayBase)),
        },
        properties={"prop": PropertyDef(AnyAttr())},
        regions=[("region", RegionDef())],
        accessor_names={"attr": ("attr", "attribute"), "prop": ("prop", "property")},
        options=[AttrSizedOperandSegments()],
    )


@irdl_op_definition
class PropOptionOp(IRDLOperation):
    name = "test.prop_option_test"

    irdl_options = [AttrSizedOperandSegments(as_property=True)]


def test_property_option():
    """Test retrieval of an IRDL definition from an operation"""
    assert PropOptionOp.get_irdl_definition() == OpDef(
        "test.prop_option_test",
        properties={"operandSegmentSizes": PropertyDef(BaseAttr(DenseArrayBase))},
        options=[AttrSizedOperandSegments(as_property=True)],
    )


################################################################################
#                            Invalid definitions                               #
################################################################################


class InvalidTypedFieldTestOp(IRDLOperation):
    name = "test.invalid_typed_field"

    field: int


def test_invalid_typed_field():
    """Check that typed fields are not allowed"""
    with pytest.raises(PyRDLOpDefinitionError):
        irdl_op_definition(InvalidTypedFieldTestOp)


class InvalidFieldTestOp(IRDLOperation):
    name = "test.invalid_field"

    field = 2


def test_invalid_field():
    """Check that untyped fields are not allowed"""
    with pytest.raises(PyRDLOpDefinitionError):
        irdl_op_definition(InvalidFieldTestOp)


################################################################################
#                                  Verifiers                                   #
################################################################################


@irdl_op_definition
class AttrOp(IRDLOperation):
    name = "test.two_var_result_op"
    attr = attr_def(StringAttr)


def test_attr_verify():
    op = AttrOp.create(attributes={"attr": IntAttr(1)})
    with pytest.raises(
        VerifyException, match="#builtin.int<1> should be of base attribute string"
    ):
        op.verify()


# TODO: remove this test once the Annotated API is deprecated
@irdl_op_definition
class ConstraintVarOp(IRDLOperation):
    name = "test.constraint_var_op"

    T = Annotated[IntegerType | IndexType, ConstraintVar("T")]

    operand = operand_def(T)  # pyright: ignore[reportArgumentType]
    result = result_def(T)  # pyright: ignore[reportArgumentType]
    attribute = attr_def(T)  # pyright: ignore[reportArgumentType, reportUnknownVariableType]


def test_constraint_var():
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())
    op = ConstraintVarOp.create(
        operands=[i32_operand], result_types=[i32], attributes={"attribute": i32}
    )
    op.verify()

    op2 = ConstraintVarOp.create(
        operands=[index_operand],
        result_types=[IndexType()],
        attributes={"attribute": IndexType()},
    )
    op2.verify()


def test_constraint_var_fail_non_equal():
    """Check that all uses of a constraint variable are of the same attribute."""
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())

    # Fail because of operand
    op = ConstraintVarOp.create(
        operands=[index_operand], result_types=[i32], attributes={"attribute": i32}
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: result at position 0 does not verify",
    ):
        op.verify()

    # Fail because of result
    op2 = ConstraintVarOp.create(
        operands=[i32_operand],
        result_types=[IndexType()],
        attributes={"attribute": i32},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: result at position 0 does not verify",
    ):
        op2.verify()

    # Fail because of attribute
    op3 = ConstraintVarOp.create(
        operands=[i32_operand],
        result_types=[i32],
        attributes={"attribute": IndexType()},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: attribute i32 expected from variable 'T', but got index",
    ):
        op3.verify()


def test_constraint_var_fail_not_satisfy_constraint():
    """Check that all uses of a constraint variable are satisfying the constraint."""
    test_operand = TestSSAValue(TestType("foo"))
    op = ConstraintVarOp.create(
        operands=[test_operand],
        result_types=[TestType("foo")],
        attributes={"attribute": TestType("foo")},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: operand at position 0 does not verify",
    ):
        op.verify()


@irdl_op_definition
class GenericConstraintVarOp(IRDLOperation):
    name = "test.constraint_var_op"

    T: ClassVar = VarConstraint("T", base(IntegerType) | base(IndexType))

    operand = operand_def(T)
    result = result_def(T)
    attribute = attr_def(T)


def test_generic_constraint_var():
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())
    op = GenericConstraintVarOp.create(
        operands=[i32_operand], result_types=[i32], attributes={"attribute": i32}
    )
    op.verify()

    op2 = GenericConstraintVarOp.create(
        operands=[index_operand],
        result_types=[IndexType()],
        attributes={"attribute": IndexType()},
    )
    op2.verify()


def test_generic_constraint_var_fail_non_equal():
    """Check that all uses of a constraint variable are of the same attribute."""
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())

    # Fail because of operand
    op = GenericConstraintVarOp.create(
        operands=[index_operand], result_types=[i32], attributes={"attribute": i32}
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: result at position 0 does not verify",
    ):
        op.verify()

    # Fail because of result
    op2 = GenericConstraintVarOp.create(
        operands=[i32_operand],
        result_types=[IndexType()],
        attributes={"attribute": i32},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: result at position 0 does not verify",
    ):
        op2.verify()

    # Fail because of attribute
    op3 = GenericConstraintVarOp.create(
        operands=[i32_operand],
        result_types=[i32],
        attributes={"attribute": IndexType()},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: attribute i32 expected from variable 'T', but got index",
    ):
        op3.verify()


def test_generic_constraint_var_fail_not_satisfy_constraint():
    """Check that all uses of a constraint variable are satisfying the constraint."""
    test_operand = TestSSAValue(TestType("foo"))
    op = GenericConstraintVarOp.create(
        operands=[test_operand],
        result_types=[TestType("foo")],
        attributes={"attribute": TestType("foo")},
    )
    with pytest.raises(
        DiagnosticException,
        match="Operation does not verify: operand at position 0 does not verify",
    ):
        op.verify()


@irdl_op_definition
class ConstraintRangeVarOp(IRDLOperation):
    name = "test.constraint_range_var"

    operand = var_operand_def(RangeVarConstraint("T", RangeOf(AnyOf((i32, IndexType)))))
    result = var_result_def(RangeVarConstraint("T", RangeOf(AnyOf((i32, IndexType)))))


def test_range_var():
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())
    op = ConstraintRangeVarOp.create(operands=[], result_types=[])
    op.verify()
    op = ConstraintRangeVarOp.create(operands=[i32_operand], result_types=[i32])
    op.verify()
    op = ConstraintRangeVarOp.create(
        operands=[i32_operand, i32_operand], result_types=[i32, i32]
    )
    op.verify()

    op2 = ConstraintRangeVarOp.create(
        operands=[index_operand], result_types=[IndexType()]
    )
    op2.verify()


def test_range_var_fail_non_equal():
    """Check that all uses of a range variable are of the same attribute."""
    i32_operand = TestSSAValue(i32)
    index_operand = TestSSAValue(IndexType())

    op = ConstraintRangeVarOp.create(operands=[index_operand], result_types=[i32])
    with pytest.raises(
        VerifyException,
        match=r"attributes \('index',\) expected from range variable 'T', but got \('i32',\)",
    ):
        op.verify()

    op2 = ConstraintRangeVarOp.create(
        operands=[i32_operand], result_types=[IndexType()]
    )
    with pytest.raises(
        VerifyException,
        match=r"attributes \('i32',\) expected from range variable 'T', but got \('index',\)",
    ):
        op2.verify()

    op2 = ConstraintRangeVarOp.create(operands=[i32_operand], result_types=[i32, i32])
    with pytest.raises(
        VerifyException,
        match=r"attributes \('i32',\) expected from range variable 'T', but got \('i32', 'i32'\)",
    ):
        op2.verify()

    op2 = ConstraintRangeVarOp.create(operands=[i32_operand], result_types=[])
    with pytest.raises(
        VerifyException,
        match=r"attributes \('i32',\) expected from range variable 'T', but got \(\)",
    ):
        op2.verify()


def test_range_var_fail_not_satisfy_constraint():
    """Check that all uses of a range variable are satisfying the constraint."""
    test_operand = TestSSAValue(TestType("foo"))
    op = ConstraintRangeVarOp.create(
        operands=[test_operand], result_types=[TestType("foo")]
    )
    with pytest.raises(VerifyException, match='Unexpected attribute !test.type<"foo">'):
        op.verify()

    op = ConstraintRangeVarOp.create(
        operands=[test_operand, test_operand],
        result_types=[TestType("foo"), TestType("foo")],
    )
    with pytest.raises(VerifyException, match='Unexpected attribute !test.type<"foo">'):
        op.verify()


@irdl_op_definition
class WithoutPropOp(IRDLOperation):
    name = "test.op_without_prop"

    prop1 = prop_def(Attribute)


# Check that an operation cannot accept properties that are not defined
def test_unknown_property():
    op = WithoutPropOp.create(properties={"prop1": i32, "prop2": i32})
    with pytest.raises(
        VerifyException, match="property 'prop2' is not defined by the operation"
    ):
        op.verify()


################################################################################
#                                Accessors                                     #
################################################################################


@irdl_op_definition
class RegionOp(IRDLOperation):
    name = "test.region_op"

    irdl_options = [AttrSizedRegionSegments()]

    region = region_def()
    opt_region = opt_region_def()
    var_region = var_region_def()


def test_region_accessors():
    """Test accessors for regions."""
    region1 = Region()
    region2 = Region()
    region3 = Region()
    region4 = Region()

    op = RegionOp.build(regions=[region1, [region2], [region3, region4]])
    assert op.region is op.regions[0]
    assert op.opt_region is op.regions[1]
    assert len(op.var_region) == 2
    assert op.var_region[0] is op.regions[2]
    assert op.var_region[1] is op.regions[3]

    region1 = Region()

    op = RegionOp.build(regions=[region1, [], []])
    assert op.opt_region is None
    assert len(op.var_region) == 0


@irdl_op_definition
class OperandOp(IRDLOperation):
    name = "test.operand_op"

    irdl_options = [AttrSizedOperandSegments()]

    operand = operand_def()
    opt_operand = opt_operand_def()
    var_operand = var_operand_def()


def test_operand_accessors():
    """Test accessors for operands."""
    operand1 = TestSSAValue(i32)
    operand2 = TestSSAValue(i32)
    operand3 = TestSSAValue(i32)
    operand4 = TestSSAValue(i32)

    op = OperandOp.build(operands=[operand1, [operand2], [operand3, operand4]])
    assert op.operand is op.operands[0]
    assert op.opt_operand is op.operands[1]
    assert len(op.var_operand) == 2
    assert op.var_operand[0] is op.operands[2]
    assert op.var_operand[1] is op.operands[3]

    op = OperandOp.build(operands=[operand1, [], []])
    assert op.opt_operand is None
    assert len(op.var_operand) == 0


@irdl_op_definition
class OpResultOp(IRDLOperation):
    name = "test.op_result_op"

    irdl_options = [AttrSizedResultSegments()]

    result = result_def()
    opt_result = opt_result_def()
    var_result = var_result_def()


def test_opresult_accessors():
    """Test accessors for results."""
    op = OpResultOp.build(result_types=[i32, [i32], [i32, i32]])
    assert op.result is op.results[0]
    assert op.opt_result is op.results[1]
    assert len(op.var_result) == 2
    assert op.var_result[0] is op.results[2]
    assert op.var_result[1] is op.results[3]

    op = OpResultOp.build(result_types=[i32, [], []])
    assert op.opt_result is None
    assert len(op.var_result) == 0


@irdl_op_definition
class AttributeOp(IRDLOperation):
    name = "test.attribute_op"

    attr = attr_def(StringAttr)
    opt_attr = opt_attr_def(StringAttr)


def test_attribute_accessors():
    """Test accessors for attributes."""

    op = AttributeOp.create(
        attributes={"attr": StringAttr("test"), "opt_attr": StringAttr("opt_test")}
    )
    assert op.attr is op.attributes["attr"]
    assert op.opt_attr is op.attributes["opt_attr"]

    op = AttributeOp.create(attributes={"attr": StringAttr("test")})
    assert op.opt_attr is None


def test_attribute_setters():
    """Test setters for attributes."""

    op = AttributeOp.create(attributes={"attr": StringAttr("test")})

    op.attr = StringAttr("new_test")
    assert op.attr.data == "new_test"

    op.opt_attr = StringAttr("new_opt_test")
    assert op.opt_attr.data == "new_opt_test"

    op.opt_attr = None
    assert op.opt_attr is None


@irdl_op_definition
class PropertyOp(IRDLOperation):
    name = "test.attribute_op"

    attr = prop_def(StringAttr)
    opt_attr = opt_prop_def(StringAttr)


def test_property_accessors():
    """Test accessors for properties."""

    op = PropertyOp.create(
        properties={"attr": StringAttr("test"), "opt_attr": StringAttr("opt_test")}
    )
    assert op.attr is op.properties["attr"]
    assert op.opt_attr is op.properties["opt_attr"]

    op = PropertyOp.create(properties={"attr": StringAttr("test")})
    assert op.opt_attr is None


def test_property_setters():
    """Test setters for properties."""

    op = PropertyOp.create(properties={"attr": StringAttr("test")})

    op.attr = StringAttr("new_test")
    assert op.attr.data == "new_test"

    op.opt_attr = StringAttr("new_opt_test")
    assert op.opt_attr.data == "new_opt_test"

    op.opt_attr = None
    assert op.opt_attr is None


################################################################################
#                            Renamed attributes                                #
################################################################################

# These tests check that operations that have attributes with names that differ
# from the attribute accessors can be defined.


@irdl_op_definition
class RenamedAttributeOp(IRDLOperation):
    """
    An operation that has attributes with different names than the attribute
    accessors.
    """

    name = "test.renamed_attribute_op"

    accessor = attr_def(StringAttr, attr_name="attr_name")
    opt_accessor = opt_attr_def(StringAttr, attr_name="opt_attr_name")


def test_renamed_attributes_verify():
    op = RenamedAttributeOp.create(
        attributes={
            "attr_name": StringAttr("test"),
            "opt_attr_name": StringAttr("test_opt"),
        }
    )
    op.verify()

    op = RenamedAttributeOp.create(
        attributes={
            "accessor": StringAttr("test"),
        }
    )
    with pytest.raises(VerifyException, match="attribute attr_name expected"):
        op.verify()

    op = RenamedAttributeOp.create(
        attributes={
            "attr_name": StringAttr("test"),
            "opt_attr_name": i32,
        }
    )
    with pytest.raises(VerifyException, match="i32 should be of base attribute string"):
        op.verify()


def test_renamed_attributes_accessors():
    op = RenamedAttributeOp.create(
        attributes={
            "attr_name": StringAttr("test"),
            "opt_attr_name": StringAttr("test_opt"),
        }
    )

    assert op.accessor is op.attributes["attr_name"]
    assert op.opt_accessor is op.attributes["opt_attr_name"]


@irdl_op_definition
class RenamedPropertyOp(IRDLOperation):
    """
    An operation that has properties with different names than the properties
    accessors.
    """

    name = "test.renamed_property_op"

    accessor = prop_def(StringAttr, prop_name="prop_name")
    opt_accessor = opt_prop_def(StringAttr, prop_name="opt_prop_name")


def test_renamed_properties_verify():
    op = RenamedPropertyOp.create(
        properties={
            "prop_name": StringAttr("test"),
            "opt_prop_name": StringAttr("test_opt"),
        }
    )
    op.verify()

    op = RenamedPropertyOp.create(
        properties={
            "accessor": StringAttr("test"),
        }
    )
    with pytest.raises(VerifyException, match="property prop_name expected"):
        op.verify()

    op = RenamedPropertyOp.create(
        properties={
            "prop_name": StringAttr("test"),
            "opt_prop_name": i32,
        }
    )
    with pytest.raises(VerifyException, match="i32 should be of base attribute string"):
        op.verify()


def test_renamed_properties_accessors():
    op = RenamedPropertyOp.create(
        properties={
            "prop_name": StringAttr("test"),
            "opt_prop_name": StringAttr("test_opt"),
        }
    )

    assert op.accessor is op.properties["prop_name"]
    assert op.opt_accessor is op.properties["opt_prop_name"]


################################################################################
#                             Generic operation                                #
################################################################################

FooType = Annotated[TestType, TestType("foo")]
BarType = Annotated[TestType, TestType("bar")]


_Attr = TypeVar("_Attr", bound=StringAttr | IntAttr)
_Operand = TypeVar("_Operand", bound=FooType | BarType)
_Result = TypeVar("_Result", bound=FooType | BarType)


class GenericOp(Generic[_Attr, _Operand, _Result], IRDLOperation):
    name = "test.string_or_int_generic"

    attr: _Attr = attr_def(_Attr)
    operand = operand_def(_Operand)
    result = result_def(_Result)


@irdl_op_definition
class StringFooOp(GenericOp[StringAttr, FooType, FooType]):
    name = "test.string_specialized"


def test_generic_op():
    """Test generic operation."""
    FooOperand = TestSSAValue(TestType("foo"))
    BarOperand = TestSSAValue(TestType("bar"))
    FooResultType = TestType("foo")
    BarResultType = TestType("bar")

    op = StringFooOp(
        attributes={"attr": StringAttr("test")},
        operands=[FooOperand],
        result_types=[FooResultType],
    )
    op.verify()

    op_attr_fail = StringFooOp(
        attributes={"attr": IntAttr(1)},
        operands=[FooOperand],
        result_types=[FooResultType],
    )
    with pytest.raises(DiagnosticException):
        op_attr_fail.verify()

    op_operand_fail = StringFooOp(
        attributes={"attr": StringAttr("test")},
        operands=[BarOperand],
        result_types=[FooResultType],
    )
    with pytest.raises(DiagnosticException):
        op_operand_fail.verify()

    op_result_fail = StringFooOp(
        attributes={"attr": StringAttr("test")},
        operands=[FooOperand],
        result_types=[BarResultType],
    )
    with pytest.raises(DiagnosticException):
        op_result_fail.verify()


class OtherParentOp(IRDLOperation):
    other_attr = attr_def(Attribute)


@irdl_op_definition
class OtherStringFooOp(GenericOp[StringAttr, FooType, FooType], OtherParentOp):
    name = "test.string_specialized"


def test_multiple_inheritance_op():
    """Test generic operation."""
    FooOperand = TestSSAValue(TestType("foo"))
    FooResultType = TestType("foo")

    op = OtherStringFooOp(
        attributes={"attr": StringAttr("test"), "other_attr": StringAttr("test")},
        operands=[FooOperand],
        result_types=[FooResultType],
    )
    op.verify()

    op_attr_fail = OtherStringFooOp(
        attributes={"attr": IntAttr(1), "other_attr": StringAttr("test")},
        operands=[FooOperand],
        result_types=[FooResultType],
    )
    with pytest.raises(DiagnosticException):
        op_attr_fail.verify()

    op = OtherStringFooOp(
        attributes={"attr": StringAttr("test")},
        operands=[FooOperand],
        result_types=[FooResultType],
    )
    with pytest.raises(DiagnosticException):
        op_attr_fail.verify()


@irdl_op_definition
class EntryArgsOp(IRDLOperation):
    name = "test.entry_args"
    body = opt_region_def(entry_args=RangeOf(EqAttrConstraint(i32)))

    traits = traits_def(NoTerminator())


def test_entry_args_op():
    op = EntryArgsOp.create()
    op.verify()

    op = EntryArgsOp.create(regions=[Region(Block(arg_types=[]))])
    op.verify()
    op = EntryArgsOp.create(regions=[Region(Block(arg_types=[i32]))])
    op.verify()
    op = EntryArgsOp.create(regions=[Region(Block(arg_types=[i32, i32]))])
    op.verify()

    op = EntryArgsOp.create(regions=[Region(Block(arg_types=[i64]))])
    with pytest.raises(
        DiagnosticException,
        match="""\
Operation does not verify: region #0 entry arguments do not verify:
Expected attribute i32 but got i64""",
    ):
        op.verify()

    op = EntryArgsOp.create(regions=[Region(Block(arg_types=[i64, i32]))])
    with pytest.raises(
        DiagnosticException,
        match="""\
Operation does not verify: region #0 entry arguments do not verify:
Expected attribute i32 but got i64""",
    ):
        op.verify()


class OptionlessMultipleVarOp(IRDLOperation):
    name = "test.multiple_var_op"

    optional = opt_operand_def()
    variadic = var_operand_def()


def test_no_multiple_var_option():
    with pytest.raises(
        PyRDLOpDefinitionError,
        match="Operation test.multiple_var_op defines more than two variadic operands, "
        "but do not define any of SameVariadicOperandSize or AttrSizedOperandSegments "
        "PyRDL options.",
    ):
        irdl_op_definition(OptionlessMultipleVarOp)


@irdl_op_definition
class DefaultOp(IRDLOperation):
    name = "test.default"

    prop = prop_def(BoolAttr, default_value=BoolAttr.from_bool(False))
    opt_prop = opt_prop_def(BoolAttr, default_value=BoolAttr.from_bool(True))

    attr = attr_def(BoolAttr, default_value=BoolAttr.from_bool(False))
    opt_attr = opt_attr_def(BoolAttr, default_value=BoolAttr.from_bool(True))

    assembly_format = "(`prop` $prop^)? (`opt_prop` $opt_prop^)? (`attr` $attr^)? (`opt_attr` $opt_attr^)? attr-dict"


def test_default_accessors():
    ctx = Context()
    ctx.load_op(DefaultOp)

    parsed = Parser(ctx, "test.default").parse_operation()

    assert isinstance(parsed, DefaultOp)

    assert not parsed.prop.value.data

    assert parsed.properties.get("opt_prop") is None

    assert parsed.opt_prop.value.data

    assert not parsed.attr.value.data

    assert parsed.attributes.get("opt_attr") is None

    assert parsed.opt_attr.value.data


def test_generic_accessors():
    ctx = Context()
    ctx.load_op(DefaultOp)

    parsed = Parser(
        ctx, '"test.default"() <{ "prop" = false }> {"attr" = false} : () -> ()'
    ).parse_operation()

    assert isinstance(parsed, DefaultOp)

    assert not parsed.prop.value.data

    assert parsed.properties.get("opt_prop") is None

    assert parsed.opt_prop.value.data

    assert not parsed.attr.value.data

    assert parsed.attributes.get("opt_attr") is None

    assert parsed.opt_attr.value.data



from xdsl.dialects import memref, mpi
from xdsl.dialects.arith import ConstantOp
from xdsl.dialects.builtin import f64, i32


def test_mpi_baseop():
    """
    This test is used to track changes in `.get` and other accessors
    """
    alloc0 = memref.AllocOp.get(f64, 32, [100, 14, 14])
    dest = ConstantOp.from_int_and_width(1, i32)
    unwrap = mpi.UnwrapMemRefOp(alloc0)
    req_vec = mpi.AllocateTypeOp(mpi.RequestType, dest)
    req_obj = mpi.VectorGetOp(req_vec, dest)
    tag = ConstantOp.from_int_and_width(1, i32)
    send = mpi.IsendOp(unwrap.ptr, unwrap.len, unwrap.type, dest, tag, req_obj)
    wait = mpi.WaitOp(send.request, ignore_status=False)
    recv = mpi.IrecvOp(unwrap.ptr, unwrap.len, unwrap.type, dest, tag, req_obj)
    test_res = mpi.TestOp(recv.request)
    assert wait.status is not None
    source = mpi.GetStatusFieldOp(wait.status, mpi.StatusTypeField.MPI_SOURCE)

    assert unwrap.ref == alloc0.memref
    assert send.buffer == unwrap.ptr
    assert send.count == unwrap.len
    assert send.datatype == unwrap.type
    assert send.dest == dest.result
    assert send.tag == tag.result
    assert wait.request == send.request
    assert recv.buffer == unwrap.ptr
    assert recv.count == unwrap.len
    assert recv.datatype == unwrap.type
    assert recv.source == dest.result
    assert recv.tag == tag.result
    assert test_res.request == recv.request
    assert source.status == wait.status
    assert source.field.data == mpi.StatusTypeField.MPI_SOURCE.value



from xdsl.utils.comparisons import (
    signed_lower_bound,
    signed_upper_bound,
    signed_value_range,
    signless_value_range,
    to_signed,
    to_unsigned,
    unsigned_upper_bound,
    unsigned_value_range,
)

BITWIDTH = 2
UNSIGNED_UPPER_BOUND = unsigned_upper_bound(BITWIDTH)
SIGNED_UPPER_BOUND = signed_upper_bound(BITWIDTH)
SIGNED_LOWER_BOUND = signed_lower_bound(BITWIDTH)

UNSIGNED_RANGE = range(UNSIGNED_UPPER_BOUND)
SIGNED_RANGE = range(SIGNED_LOWER_BOUND, SIGNED_UPPER_BOUND)


def unsigned_to_signed(u: int) -> int:
    return u + SIGNED_LOWER_BOUND


def test_bitwidth_2_values():
    """
    Above calculations are correct for bitwidth 2.
    """
    assert list(SIGNED_RANGE) == [unsigned_to_signed(u) for u in UNSIGNED_RANGE]


def test_bitwidth_0_values():
    """
    We support 0-width types, where the value is always 0.
    """
    assert unsigned_value_range(0) == (0, 1)
    assert signed_value_range(0) == (0, 1)
    assert signless_value_range(0) == (0, 1)


def test_conversion():
    assert to_unsigned(SIGNED_LOWER_BOUND, BITWIDTH) == SIGNED_UPPER_BOUND
    assert to_unsigned(-1, BITWIDTH) == UNSIGNED_UPPER_BOUND - 1
    assert to_unsigned(0, BITWIDTH) == 0
    assert to_unsigned(1, BITWIDTH) == 1
    assert to_unsigned(SIGNED_UPPER_BOUND, BITWIDTH) == SIGNED_UPPER_BOUND
    assert to_unsigned(UNSIGNED_UPPER_BOUND, BITWIDTH) == 0

    assert to_signed(SIGNED_LOWER_BOUND, BITWIDTH) == SIGNED_LOWER_BOUND
    assert to_signed(-1, BITWIDTH) == -1
    assert to_signed(0, BITWIDTH) == 0
    assert to_signed(1, BITWIDTH) == 1
    assert to_signed(SIGNED_UPPER_BOUND, BITWIDTH) == SIGNED_LOWER_BOUND
    assert to_signed(UNSIGNED_UPPER_BOUND, BITWIDTH) == 0



from __future__ import annotations

from collections.abc import Callable, Generator, Iterable
from dataclasses import dataclass
from enum import Enum

from .location import Location

INDENT = 2


@dataclass
class VarType:
    "A variable type with shape information."

    shape: list[int]


class ExprASTKind(Enum):
    Expr_VarDecl = 1
    Expr_Return = 2
    Expr_Num = 3
    Expr_Literal = 4
    Expr_Var = 5
    Expr_BinOp = 6
    Expr_Call = 7
    Expr_Print = 8


@dataclass()
class Dumper:
    lines: list[str]
    indentation: int = 0

    def append(self, prefix: str, line: str):
        self.lines.append(" " * self.indentation * INDENT + prefix + line)

    def append_list(
        self,
        prefix: str,
        open_paren: str,
        exprs: Iterable[ExprAST | FunctionAST],
        close_paren: str,
        block: Callable[[Dumper, ExprAST | FunctionAST], None],
    ):
        self.append(prefix, open_paren)
        child = self.child()
        for expr in exprs:
            block(child, expr)
        self.append("", close_paren)

    def child(self):
        return Dumper(self.lines, self.indentation + 1)

    @property
    def message(self):
        return "\n".join(self.lines)


@dataclass
class ExprAST:
    loc: Location

    def __init__(self, loc: Location):
        self.loc = loc
        print(self.dump())

    @property
    def kind(self) -> ExprASTKind:
        raise AssertionError(f"ExprAST kind not defined for {type(self)}")

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, self.__class__.__name__)

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message


@dataclass
class VarDeclExprAST(ExprAST):
    "Expression class for defining a variable."

    name: str
    varType: VarType
    expr: ExprAST

    @property
    def kind(self):
        return ExprASTKind.Expr_VarDecl

    def inner_dump(self, prefix: str, dumper: Dumper):
        dims_str = ", ".join(f"{int(dim)}" for dim in self.varType.shape)
        dumper.append("VarDecl ", f"{self.name}<{dims_str}> @{self.loc}")
        child = dumper.child()
        self.expr.inner_dump("", child)


@dataclass
class ReturnExprAST(ExprAST):
    "Expression class for a return operator."

    expr: ExprAST | None

    @property
    def kind(self):
        return ExprASTKind.Expr_Return

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, "Return")
        if self.expr is not None:
            child = dumper.child()
            self.expr.inner_dump("", child)


@dataclass
class NumberExprAST(ExprAST):
    'Expression class for numeric literals like "1.0".'

    val: float

    @property
    def kind(self):
        return ExprASTKind.Expr_Num

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, f" {self.val:.6e}")


@dataclass
class LiteralExprAST(ExprAST):
    "Expression class for a literal value."

    values: list[LiteralExprAST | NumberExprAST]
    dims: list[int]

    @property
    def kind(self):
        return ExprASTKind.Expr_Literal

    def __dump(self) -> str:
        dims_str = ", ".join(f"{int(dim)}" for dim in self.dims)
        vals_str = ",".join(
            val.__dump() if isinstance(val, LiteralExprAST) else val.dump()
            for val in self.values
        )
        return f" <{dims_str}>[{vals_str}]"

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("Literal:", self.__dump() + f" @{self.loc}")

    def iter_flattened_values(self) -> Generator[float, None, None]:
        for value in self.values:
            if isinstance(value, NumberExprAST):
                yield value.val
            else:
                yield from value.iter_flattened_values()

    def flattened_values(self) -> list[float]:
        return list(self.iter_flattened_values())


@dataclass
class VariableExprAST(ExprAST):
    'Expression class for referencing a variable, like "a".'

    name: str

    @property
    def kind(self):
        return ExprASTKind.Expr_Var

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("var: ", f"{self.name} @{self.loc}")


@dataclass
class BinaryExprAST(ExprAST):
    "Expression class for a binary operator."

    op: str
    lhs: ExprAST
    rhs: ExprAST

    @property
    def kind(self):
        return ExprASTKind.Expr_BinOp

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, f"BinOp: {self.op} @{self.loc}")
        child = dumper.child()
        self.lhs.inner_dump("", child)
        self.rhs.inner_dump("", child)


@dataclass
class CallExprAST(ExprAST):
    "Expression class for function calls."

    callee: str
    args: list[ExprAST]

    @property
    def kind(self):
        return ExprASTKind.Expr_Call

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append_list(
            prefix,
            f"Call '{self.callee}' [ @{self.loc}",
            self.args,
            "]",
            lambda dd, arg: arg.inner_dump("", dd),
        )


@dataclass
class PrintExprAST(ExprAST):
    "Expression class for builtin print calls."

    arg: ExprAST

    @property
    def kind(self):
        return ExprASTKind.Expr_Print

    def inner_dump(self, prefix: str, dumper: Dumper):
        super().inner_dump(prefix, dumper)
        child = dumper.child()
        self.arg.inner_dump("arg: ", child)


@dataclass
class PrototypeAST:
    """
    This class represents the "prototype" for a function, which captures its
    name, and its argument names (thus implicitly the number of arguments the
    function takes).
    """

    loc: Location
    name: str
    args: list[VariableExprAST]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append("", f"Proto '{self.name}' @{self.loc}")
        dumper.append("Params: ", f"[{', '.join(arg.name for arg in self.args)}]")


@dataclass
class FunctionAST:
    "This class represents a function definition itself."

    loc: Location
    proto: PrototypeAST
    body: tuple[ExprAST, ...]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append(prefix, "Function ")
        child = dumper.child()
        self.proto.inner_dump("proto: ", child)
        child.append_list(
            "Block ",
            "{",
            self.body,
            "} // Block",
            lambda dd, stmt: stmt.inner_dump("", dd),
        )


@dataclass
class ModuleAST:
    "This class represents a list of functions to be processed together"

    funcs: tuple[FunctionAST, ...]

    def dump(self):
        dumper = Dumper([])
        self.inner_dump("", dumper)
        return dumper.message

    def inner_dump(self, prefix: str, dumper: Dumper):
        dumper.append_list(
            prefix, "Module:", self.funcs, "", lambda dd, func: func.inner_dump("", dd)
        )



import pytest

from xdsl.builder import Builder, ImplicitBuilder
from xdsl.dialects import arith, cf, func
from xdsl.dialects.builtin import ModuleOp, i32
from xdsl.interpreter import Interpreter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.cf import CfFunctions
from xdsl.interpreters.func import FuncFunctions
from xdsl.ir import Block, Region


def sum_to_fn(n: int) -> int:
    result = 0
    i = 0
    while i <= n:
        result += i
        i += 1
    return result


@ModuleOp
@Builder.implicit_region
def sum_to_op():
    prologue = Block(arg_types=(i32,))
    loop_iter = Block(arg_types=(i32, i32))
    loop_body = Block(arg_types=(i32, i32))
    epilogue = Block(arg_types=(i32,))

    with ImplicitBuilder(prologue):
        result = arith.ConstantOp.from_int_and_width(0, 32).result
        i = arith.ConstantOp.from_int_and_width(0, 32).result
        cf.BranchOp(loop_iter, result, i)

    with ImplicitBuilder(loop_iter):
        (n,) = prologue.args
        result, i = loop_iter.args
        cond = arith.CmpiOp(i, n, "sle")
        cf.ConditionalBranchOp(cond, loop_body, (result, i), epilogue, (result,))

    with ImplicitBuilder(loop_body):
        result, i = loop_iter.args
        new_result = arith.AddiOp(result, i)
        one = arith.ConstantOp.from_int_and_width(1, 32)
        new_i = arith.AddiOp(i, one)
        cf.BranchOp(loop_iter, new_result, new_i)

    with ImplicitBuilder(epilogue):
        (result,) = epilogue.args
        func.ReturnOp(result)

    func.FuncOp(
        "sum_to", ((i32,), (i32,)), Region([prologue, loop_iter, loop_body, epilogue])
    )


def sum_to_interp(n: int) -> int:
    interpreter = Interpreter(sum_to_op)
    interpreter.register_implementations(CfFunctions())
    interpreter.register_implementations(FuncFunctions())
    interpreter.register_implementations(ArithFunctions())
    (result,) = interpreter.call_op("sum_to", (n,))
    return result


@pytest.mark.parametrize("n", [0, 1, 2, 3, 4])
def test_sum_to(n: int):
    assert sum_to_fn(n) == sum_to_interp(n)



from dataclasses import dataclass, field

from xdsl.ir import Operation
from xdsl.traits import HasInsnRepresentation


@dataclass(frozen=True)
class StaticInsnRepresentation(HasInsnRepresentation):
    """
    Returns the first parameter as an insn template string.

    See external [documentation](https://sourceware.org/binutils/docs/as/RISC_002dV_002dDirectives.html).
    """

    insn: str = field(kw_only=True)

    def get_insn(self, op: Operation) -> str:
        """
        Return the insn representation of the operation for printing.
        """
        return self.insn
